@@ -5,16 +5,16 @@
 
 #include "earley_parser.h"
 
+#include <algorithm>
 #include <cassert>
 #include <cctype>
 #include <cstdint>
 #include <ctime>
-#include <unordered_map>
 #include <utility>
 #include <vector>
 
 #include "fsm.h"
-#include "grammar_data_structure.h"
+#include "grammar_impl.h"
 #include "support/encoding.h"
 #include "support/logging.h"
 #include "xgrammar/grammar.h"
@@ -31,97 +31,146 @@ void EarleyParser::PopLastStates(int32_t cnt) {
   if (stop_token_is_accepted_) {
     stop_token_is_accepted_ = false;
   }
-  if (cnt >= static_cast<int32_t>(rule_id_to_completeable_states_.size())) {
+  if (cnt >= static_cast<int32_t>(rule_id_to_completable_states_.size())) {
     XGRAMMAR_LOG(FATAL) << "The number of states to be popped is larger than the size of states.";
   }
-  rule_id_to_completeable_states_.erase(
-      rule_id_to_completeable_states_.end() - cnt, rule_id_to_completeable_states_.end()
-  );
+  rule_id_to_completable_states_.PopBack(cnt);
   is_completed_.erase(is_completed_.end() - cnt, is_completed_.end());
   scanable_state_history_.PopBack(cnt);
 }
 
-void EarleyParser::Complete(const ParserState& state, const GrammarExpr& grammar_expr) {
+void EarleyParser::Complete(const ParserState& state, bool debug_print) {
   // Check if a rule is completed.
   if (state.rule_start_pos == ParserState::kNoPrevInputPos) {
     // assert: if a root rule can achieve here, then it must be completed.
-    XGRAMMAR_DCHECK(grammar_expr.type == GrammarExprType::kSequence);
-    XGRAMMAR_DCHECK(grammar_expr.size() == state.element_id);
+    if (debug_print) {
+      XGRAMMAR_LOG(INFO) << "The root rule is completed.";
+    }
     tmp_accept_stop_token_ = true;
     return;
   }
+  if (debug_print) {
+    XGRAMMAR_LOG(INFO) << "The rule " << state.rule_id << ": "
+                       << grammar_->GetRule(state.rule_id).name
+                       << " is completed, trying to complete its parent states.";
+  }
+
   // Check all the possible parent states.
-  const auto& parent_states_map = rule_id_to_completeable_states_[state.rule_start_pos];
-  for (auto parent_state_iter = parent_states_map.lower_bound(state.rule_id);
-       parent_state_iter != parent_states_map.end() && parent_state_iter->first == state.rule_id;
-       parent_state_iter++) {
-    const auto& parent_state = parent_state_iter->second;
-    const auto& parent_expr = grammar_->GetGrammarExpr(parent_state.sequence_id);
-    if (parent_expr.type == GrammarExprType::kSequence) {
-      // These two types can predict other new rules. We need to
-      // to move to the next element.
-      XGRAMMAR_DCHECK(
-          grammar_->GetGrammarExpr(parent_expr[parent_state.element_id]).type ==
-          GrammarExprType::kRuleRef
-      );
-      Enqueue(ParserState{
-          parent_state.rule_id,
-          parent_state.sequence_id,
-          parent_state.element_id + 1,
-          parent_state.rule_start_pos,
-          0
-      });
+  const auto& parent_states_map = rule_id_to_completable_states_[state.rule_start_pos];
+  for (const auto& [ref_id, parent_state] : parent_states_map) {
+    if (ref_id != state.rule_id) {
       continue;
     }
-    XGRAMMAR_DCHECK(parent_expr.type == GrammarExprType::kTagDispatch);
-    Enqueue(
-        {parent_state.rule_id,
-         parent_state.sequence_id,
-         grammar_->root_tag_dispatch_fsm->GetStart(),
-         parent_state.rule_start_pos,
-         0}
-    );
+    if (parent_state.rule_id == -1 || !grammar_->per_rule_fsms[parent_state.rule_id].has_value()) {
+      const auto& parent_expr = grammar_->GetGrammarExpr(parent_state.sequence_id);
+      const auto& element_expr = grammar_->GetGrammarExpr(parent_expr[parent_state.element_id]);
+      // The new rule is not referenced by a fsm.
+      XGRAMMAR_DCHECK(
+          element_expr.type == GrammarExprType::kRuleRef ||
+          element_expr.type == GrammarExprType::kRepeat
+      );
+      if (element_expr.type == GrammarExprType::kRuleRef) {
+        Enqueue(ParserState{
+            parent_state.rule_id,
+            parent_state.sequence_id,
+            parent_state.element_id + 1,
+            parent_state.rule_start_pos,
+            0
+        });
+        continue;
+      }
+      XGRAMMAR_DCHECK(element_expr.type == GrammarExprType::kRepeat);
+      // The parent state is a repeat, we need to increase the repeat count.
+      auto new_state = parent_state;
+      const int32_t& min_repeat_count = element_expr[1];
+      const int32_t& max_repeat_count = element_expr[2];
+      new_state.repeat_count++;
+      // The repeat rule can be completed, and we advance the state. Don't forget to
+      // reset the repeat count.
+      if (new_state.repeat_count >= min_repeat_count) {
+        Enqueue(ParserState{
+            parent_state.rule_id,
+            parent_state.sequence_id,
+            parent_state.element_id + 1,
+            parent_state.rule_start_pos,
+            0
+        });
+      }
+      // If the repeat count is less than the max repeat count, we can continue to
+      // visit the repeat state for another round.
+      if (new_state.repeat_count < max_repeat_count) {
+        Enqueue(new_state);
+      }
+      continue;
+    }
+    // If the rule is referenced by a fsm, we need to advance the fsm.
+    XGRAMMAR_DCHECK(grammar_->per_rule_fsms[parent_state.rule_id].has_value());
+    Enqueue(parent_state);
   }
 }
 
 std::pair</* scanable */ bool, /* completable */ bool> EarleyParser::Predict(
-    const ParserState& state, const GrammarExpr& grammar_expr
+    const ParserState& state, bool debug_print
 ) {
-  //  If the current state is the end of the rule, we do not need to predict.
-  if (grammar_expr.type == GrammarExprType::kTagDispatch) {
-    // The rule can be scanned, but can't be completed.
-    if (!grammar_->root_tag_dispatch_fsm->IsEndState(state.element_id)) {
-      tmp_accept_stop_token_ = true;
-      return std::make_pair(true, false);
-    }
-    // A tag has is dispatched.
-    ExpandNextRuleRefElement(state, grammar_expr, nullptr);
-    return std::make_pair(false, false);
+  // Check if the rule has a corresponding FSM.
+  if (state.rule_id != -1 && grammar_->per_rule_fsms[state.rule_id].has_value()) {
+    // Try to expand the fsm.
+    ExpandNextRuleRefElementOnFSM(state, debug_print);
+    const auto& fsm = grammar_->per_rule_fsms[state.rule_id].value();
+    return std::make_pair(fsm.IsScanableState(state.element_id), fsm.IsEndState(state.element_id));
   }
-  XGRAMMAR_DCHECK(grammar_expr.type == GrammarExprType::kSequence);
+  const GrammarExpr& grammar_expr = grammar_->GetGrammarExpr(state.sequence_id);
+  XGRAMMAR_DCHECK(
+      grammar_expr.type == GrammarExprType::kSequence ||
+      grammar_expr.type == GrammarExprType::kEmptyStr
+  );
   if (state.element_id == grammar_expr.size()) {
     // The rule is completed.
     return std::make_pair(false, true);
   }
   const auto& element_expr = grammar_->GetGrammarExpr(grammar_expr[state.element_id]);
-  if (element_expr.type == GrammarExprType::kRuleRef) {
-    ExpandNextRuleRefElement(state, grammar_expr, &element_expr);
-    return std::make_pair(false, false);
+  switch (element_expr.type) {
+    case GrammarExprType::kRuleRef: {
+      ExpandNextRuleRefElement(state, grammar_expr, &element_expr, debug_print);
+      return std::make_pair(false, false);
+    }
+    case GrammarExprType::kCharacterClassStar: {
+      if (state.sub_element_id == 0) {
+        Enqueue(ParserState{
+            state.rule_id, state.sequence_id, state.element_id + 1, state.rule_start_pos, 0
+        });
+      }
+      return std::make_pair(true, false);
+    }
+    case GrammarExprType::kRepeat: {
+      const int32_t& min_repeat_count = element_expr[1];
+      const int32_t& max_repeat_count = element_expr[2];
+      // If the current repeat count is less than the max repeat count,
+      // we can expand the next rule reference element.
+      XGRAMMAR_DCHECK(state.repeat_count <= max_repeat_count);
+      ExpandNextRuleRefElement(state, grammar_expr, &element_expr, debug_print);
+      if (state.repeat_count >= min_repeat_count) {
+        Enqueue(ParserState{
+            state.rule_id, state.sequence_id, state.element_id + 1, state.rule_start_pos, 0
+        });
+      }
+      return std::make_pair(false, false);
+    }
+    case GrammarExprType::kByteString:
+    case GrammarExprType::kCharacterClass: {
+      return std::make_pair(true, false);  // The element is scanable, but not completable.
+    }
+    default: {
+      XGRAMMAR_LOG(FATAL) << "The element type is not supported! The type is: "
+                          << int(element_expr.type);
+      XGRAMMAR_UNREACHABLE();
+    }
   }
-  if (element_expr.type == GrammarExprType::kCharacterClassStar && state.sub_element_id == 0) {
-    Enqueue(
-        ParserState{state.rule_id, state.sequence_id, state.element_id + 1, state.rule_start_pos, 0}
-    );
-  }
-  return std::make_pair(true, false);
 }
 
 void EarleyParser::Scan(const ParserState& state, const uint8_t ch) {
-  const auto& cur_rule = grammar_->GetGrammarExpr(state.sequence_id);
-  XGRAMMAR_DCHECK(
-      state.element_id != cur_rule.size() || cur_rule.type == GrammarExprType::kTagDispatch
-  );
-  if (cur_rule.type == GrammarExprType::kSequence) {
+  if (state.rule_id == -1 || (!grammar_->per_rule_fsms[state.rule_id].has_value())) {
+    const auto& cur_rule = grammar_->GetGrammarExpr(state.sequence_id);
     const auto& element_expr = grammar_->GetGrammarExpr(cur_rule[state.element_id]);
     // The element is a rule reference, we do not need to scan it.
     switch (element_expr.type) {
@@ -140,11 +189,11 @@ void EarleyParser::Scan(const ParserState& state, const uint8_t ch) {
       default: {
         XGRAMMAR_LOG(FATAL) << "The element type is not supported! The type is: "
                             << int(element_expr.type);
+        XGRAMMAR_UNREACHABLE();
       }
     }
   } else {
-    XGRAMMAR_DCHECK(cur_rule.type == GrammarExprType::kTagDispatch);
-    AdvanceTagDispatch(state, ch, cur_rule);
+    AdvanceFsm(state, ch);
   }
 }
 
@@ -159,7 +208,7 @@ void EarleyParser::Scan(const ParserState& state, const uint8_t ch) {
   \note Thus, when initializing the Earley parser, we need to add the initial state
   to the history_states[0], and perform prediction and completion on the initial state.
 */
-bool EarleyParser::Advance(const uint8_t ch) {
+bool EarleyParser::Advance(const uint8_t ch, bool debug_print) {
   // Initialize the containers.
   XGRAMMAR_DCHECK(tmp_process_state_queue_.empty())
       << "The tmp_process_state_queue_ should be empty before the scan.";
@@ -167,7 +216,6 @@ bool EarleyParser::Advance(const uint8_t ch) {
   tmp_states_to_be_added_.clear();
   tmp_accept_stop_token_ = false;
   const auto& latest_states = scanable_state_history_[scanable_state_history_.size() - 1];
-
   // Scan all the scanable states.
   for (const auto& state : latest_states) {
     Scan(state, ch);
@@ -179,15 +227,15 @@ bool EarleyParser::Advance(const uint8_t ch) {
   }
 
   // execute Predict and Complete for all states in the queue until empty.
-  rule_id_to_completeable_states_.emplace_back();
+  rule_id_to_completable_states_.PushBack(std::vector<std::pair<int32_t, ParserState>>());
   while (!tmp_process_state_queue_.empty()) {
-    const auto state = tmp_process_state_queue_.front();
+    const auto state = std::move(tmp_process_state_queue_.front());
     tmp_process_state_queue_.pop();
-    GrammarExpr grammar_expr = grammar_->GetGrammarExpr(state.sequence_id);
-    auto [scanable, completable] = Predict(state, grammar_expr);
+    auto [scanable, completable] = Predict(state, debug_print);
     if (completable) {
-      Complete(state, grammar_expr);
-    } else if (scanable) {  // A completable state can be scanned.
+      Complete(state, debug_print);
+    }
+    if (scanable) {
       tmp_states_to_be_added_.push_back(state);
     }
   }
@@ -202,6 +250,10 @@ EarleyParser::EarleyParser(
     const Grammar& grammar, const ParserState& init_state, const bool need_expand
 )
     : grammar_(grammar) {
+  if (!grammar->optimized) {
+    XGRAMMAR_LOG(FATAL) << "The grammar is not optimized. Please optimize the grammar before using "
+                           "the Earley parser.";
+  }
   // Check if the initial state is valid. If invalid, then we choose the root state as default.
   ParserState init = init_state;
   if (init_state.IsInvalid()) {
@@ -219,9 +271,10 @@ EarleyParser::EarleyParser(
   // If there is no need to expand the initial state, we only need to add it to the
   // scanable states history.
   if (!need_expand) {
-    rule_id_to_completeable_states_.emplace_back();
+    rule_id_to_completable_states_.PushBack(std::vector<std::pair<int32_t, ParserState>>());
     is_completed_.push_back(false);
     scanable_state_history_.PushBack({init});
+    return;
   }
 
   // Otherwise, we expand the initial state, and process the queue.
@@ -232,28 +285,17 @@ void EarleyParser::PushStateAndExpand(const ParserState& state) {
   tmp_states_visited_in_queue_.Clear();
   tmp_accept_stop_token_ = false;
   tmp_states_to_be_added_.clear();
-  rule_id_to_completeable_states_.emplace_back();
-  if (state.IsInvalid()) {
-    ExpandAndEnqueueUnexpandedState(ParserState{
-        grammar_->GetRootRuleId(),
-        ParserState::kUnexpandedRuleStartSequenceId,
-        0,
-        ParserState::kNoPrevInputPos,
-        0
-    });
-  } else {
-    // If the rule can't be expanded, we need to add it to the queue.
-    if (!ExpandAndEnqueueUnexpandedState(state)) {
-      Enqueue(state);
-    }
+  // If the rule can't be expanded, we need to add it to the queue.
+  if (!ExpandAndEnqueueUnexpandedState(state)) {
+    Enqueue(state);
   }
+  rule_id_to_completable_states_.PushBack(std::vector<std::pair<int32_t, ParserState>>());
   while (!tmp_process_state_queue_.empty()) {
     const auto state = tmp_process_state_queue_.front();
     tmp_process_state_queue_.pop();
-    GrammarExpr grammar_expr = grammar_->GetGrammarExpr(state.sequence_id);
-    auto [scanable, completable] = Predict(state, grammar_expr);
+    auto [scanable, completable] = Predict(state);
     if (completable) {
-      Complete(state, grammar_expr);
+      Complete(state);
     }
     if (scanable) {
       tmp_states_to_be_added_.push_back(state);
@@ -264,7 +306,7 @@ void EarleyParser::PushStateAndExpand(const ParserState& state) {
 }
 
 void EarleyParser::Reset() {
-  rule_id_to_completeable_states_.clear();
+  rule_id_to_completable_states_.PopBack(rule_id_to_completable_states_.size());
   scanable_state_history_.PopBack(scanable_state_history_.size());
   is_completed_.clear();
   stop_token_is_accepted_ = false;
@@ -289,11 +331,11 @@ bool EarleyParser::ExpandAndEnqueueUnexpandedState(const ParserState& state) {
   // There are two types of an unexpanded rule:
   // 1. The rule is a tag dispatch rule.
   // 2. The rule is a choice, consisting of multiple sequences.
-  if (cur_rule_body.type == GrammarExprType::kTagDispatch) {
+  if (state.rule_id != -1 && grammar_->per_rule_fsms[state.rule_id].has_value()) {
     Enqueue(ParserState{
         cur_rule_id,
         cur_rule_body_id,
-        grammar_->root_tag_dispatch_fsm->GetStart(),
+        grammar_->per_rule_fsms[state.rule_id]->GetStart(),
         ParserState::kNoPrevInputPos,
         0
     });
@@ -307,77 +349,98 @@ bool EarleyParser::ExpandAndEnqueueUnexpandedState(const ParserState& state) {
 }
 
 void EarleyParser::ExpandNextRuleRefElement(
-    const ParserState& state, const GrammarExpr& grammar_expr, const GrammarExpr* sub_grammar_expr
+    const ParserState& state,
+    const GrammarExpr& grammar_expr,
+    const GrammarExpr* sub_grammar_expr,
+    bool debug_print
 ) {
-  // Get the reference rule id.
-  int ref_rule_id;
-  if (grammar_expr.type == GrammarExprType::kTagDispatch) {
-    XGRAMMAR_DCHECK(grammar_->root_tag_dispatch_fsm->IsEndState(state.element_id));
-    ref_rule_id = grammar_->tag_dispatch_end_node_to_rule_id[state.element_id];
-  } else {
-    XGRAMMAR_DCHECK(grammar_expr.type == GrammarExprType::kSequence);
-    XGRAMMAR_DCHECK(sub_grammar_expr->type == GrammarExprType::kRuleRef);
-    ref_rule_id = (*sub_grammar_expr)[0];
+  // Path A. The rule has a corresponding FSM.
+  XGRAMMAR_DCHECK(!(state.rule_id != -1 && grammar_->per_rule_fsms[state.rule_id].has_value()));
+  XGRAMMAR_DCHECK(grammar_expr.type == GrammarExprType::kSequence);
+  XGRAMMAR_DCHECK(
+      sub_grammar_expr->type == GrammarExprType::kRuleRef ||
+      sub_grammar_expr->type == GrammarExprType::kRepeat
+  );
+  auto ref_rule_id = (*sub_grammar_expr)[0];
+
+  if (debug_print) {
+    XGRAMMAR_LOG(INFO) << "The rule " << state.rule_id << ": "
+                       << grammar_->GetRule(state.rule_id).name << " predict the new rule "
+                       << ref_rule_id << ": " << grammar_->GetRule(ref_rule_id).name << ".";
   }
 
-  // Add the reference rule to map.
-  if ((state.element_id != grammar_expr.size() - 1) ||
-      state.rule_start_pos == ParserState::kNoPrevInputPos) {
+  bool right_recursion_to_root = false;
+  if (state.element_id != grammar_expr.size() - 1 ||
+      sub_grammar_expr->type == GrammarExprType::kRepeat ||
+      (state.rule_start_pos == rule_id_to_completable_states_.size() - 1)) {
     // It's not the right recursion, or it's the root rule.
-    auto& states_map = rule_id_to_completeable_states_.back();
-    states_map.insert({ref_rule_id, state});
+    rule_id_to_completable_states_.PushBackInLatestRow(std::make_pair(ref_rule_id, state));
   } else {
-    // If it's the right recursion, we need to add the ancestors of the parent state.
-    auto& states_map = rule_id_to_completeable_states_.back();
-    auto& parent_states_map = rule_id_to_completeable_states_[state.rule_start_pos];
-    const auto& range = states_map.equal_range(ref_rule_id);
-    const auto in_vec = [&](const ParserState& state_) {
-      return std::find_if(range.first, range.second, [&](const auto& s) {
-               return StateEqualForParsing()(s.second, state_);
-             }) != range.second;
-    };
-    for (auto parent_state_iter = parent_states_map.lower_bound(state.rule_id);
-         parent_state_iter != parent_states_map.end() && parent_state_iter->first == state.rule_id;
-         parent_state_iter++) {
-      const auto& parent_state = parent_state_iter->second;
-      if (!in_vec(parent_state)) {
-        states_map.insert({ref_rule_id, parent_state});
+    if (state.rule_start_pos == ParserState::kNoPrevInputPos) {
+      right_recursion_to_root = true;
+    } else {
+      // If it's the right recursion, we need to add the ancestors of the parent state.
+      const auto in_vec = [&](const ParserState& state_) {
+        return std::find_if(
+                   rule_id_to_completable_states_.Back().begin(),
+                   rule_id_to_completable_states_.Back().end(),
+                   [&](const auto& s) {
+                     return StateEqualForParsing()(s.second, state_) && s.first == ref_rule_id;
+                   }
+               ) != rule_id_to_completable_states_.Back().end();
+      };
+      const auto& parent_states_map = rule_id_to_completable_states_[state.rule_start_pos];
+      std::vector<std::pair<int32_t, ParserState>> to_added_states;
+      for (const auto& parent_state_iter : parent_states_map) {
+        if (parent_state_iter.first != state.rule_id) continue;
+        const auto& parent_state = parent_state_iter.second;
+        if (!in_vec(parent_state)) {
+          to_added_states.push_back({ref_rule_id, parent_state});
+        }
+      }
+      for (const auto& to_add_state : to_added_states) {
+        rule_id_to_completable_states_.PushBackInLatestRow(to_add_state);
       }
     }
   }
 
-  // Check if the reference rule is already visited.
-  if (IsStateVisitedInQueue({ref_rule_id, -1, -1, -1, -1})) {
+  if (std::find(
+          grammar_->allow_empty_rule_ids.begin(), grammar_->allow_empty_rule_ids.end(), ref_rule_id
+      ) != grammar_->allow_empty_rule_ids.end()) {
+    XGRAMMAR_DCHECK(grammar_expr.type == GrammarExprType::kSequence);
+    Enqueue(
+        ParserState{state.rule_id, state.sequence_id, state.element_id + 1, state.rule_start_pos, 0}
+    );
+  }
+
+  // If the reference rule is not visited, we need to add it to the queue.
+  const auto& ref_rule = grammar_->GetRule(ref_rule_id);
+  const auto& ref_grammar_expr_id = ref_rule.body_expr_id;
+
+  if (grammar_->per_rule_fsms[ref_rule_id].has_value()) {
     if (std::find(
             grammar_->allow_empty_rule_ids.begin(),
             grammar_->allow_empty_rule_ids.end(),
             ref_rule_id
         ) != grammar_->allow_empty_rule_ids.end()) {
-      if (grammar_expr.type == GrammarExprType::kTagDispatch) {
-        Enqueue(ParserState{
-            state.rule_id,
-            state.sequence_id,
-            grammar_->root_tag_dispatch_fsm->GetStart(),
-            state.rule_start_pos,
-            0
-        });
-        tmp_accept_stop_token_ = true;
-        return;
-      }
-      XGRAMMAR_DCHECK(grammar_expr.type == GrammarExprType::kSequence);
       Enqueue(ParserState{
           state.rule_id, state.sequence_id, state.element_id + 1, state.rule_start_pos, 0
       });
     }
+    const auto& ref_fsm = grammar_->per_rule_fsms[ref_rule_id].value();
+    Enqueue(ParserState{
+        ref_rule_id,
+        ref_grammar_expr_id,
+        ref_fsm.GetStart(),
+        right_recursion_to_root ? ParserState::kNoPrevInputPos
+                                : int32_t(rule_id_to_completable_states_.size() - 1),
+        0
+    });
     return;
   }
 
-  // If the reference rule is not visited, we need to add it to the queue.
-  tmp_states_visited_in_queue_.Insert({ref_rule_id, -1, -1, -1, -1});
-  const auto& ref_rule = grammar_->GetRule(ref_rule_id);
-  const auto& ref_grammar_expr_id = ref_rule.body_expr_id;
   const auto& ref_grammar_expr = grammar_->GetGrammarExpr(ref_grammar_expr_id);
-  XGRAMMAR_DCHECK(ref_grammar_expr.type == GrammarExprType::kChoices);
+  XGRAMMAR_DCHECK(!grammar_->per_rule_fsms[ref_rule_id].has_value());
   for (const auto& sequence_id : ref_grammar_expr) {
     const auto& sequence = grammar_->GetGrammarExpr(sequence_id);
     if (sequence.type == GrammarExprType::kEmptyStr) {
@@ -386,14 +449,127 @@ void EarleyParser::ExpandNextRuleRefElement(
       });
       continue;
     }
-    // Assert: the state can't be repeated. Since the rule_start_pos is the current
-    // position, and the rule can only be predicted once.
-    tmp_process_state_queue_.push(ParserState{
-        ref_rule_id, sequence_id, 0, int32_t(rule_id_to_completeable_states_.size()) - 1, 0
+    Enqueue(ParserState{
+        ref_rule_id,
+        sequence_id,
+        0,
+        right_recursion_to_root ? ParserState::kNoPrevInputPos
+                                : int32_t(rule_id_to_completable_states_.size() - 1),
+        0
     });
   }
 }
 
+void EarleyParser::ExpandNextRuleRefElementOnFSM(const ParserState& state, bool debug_print) {
+  XGRAMMAR_DCHECK(state.rule_id != -1 && grammar_->per_rule_fsms[state.rule_id].has_value());
+  const auto& fsm = grammar_->per_rule_fsms[state.rule_id].value();
+
+  // Add the rule reference pairs, and enqueue the epsilon edges.
+  for (const auto& edge : fsm.GetFsm().GetEdges(state.element_id)) {
+    if (edge.IsEpsilon()) {
+      Enqueue(ParserState{state.rule_id, state.sequence_id, edge.target, state.rule_start_pos, 0});
+      continue;
+    }
+    if (!edge.IsRuleRef()) {
+      continue;
+    }
+    const int& target = edge.target;
+    const int& ref_rule_id = edge.GetRefRuleId();
+    bool right_recursion_to_root = false;
+    if (debug_print) {
+      XGRAMMAR_LOG(INFO) << "The rule " << state.rule_id << ": "
+                         << grammar_->GetRule(state.rule_id).name << " predict the new rule "
+                         << ref_rule_id << ": " << grammar_->GetRule(ref_rule_id).name << ".";
+    }
+    if ((fsm.GetFsm().GetEdges(target).size() == 0) && fsm.IsEndState(target) &&
+        state.rule_start_pos != static_cast<int32_t>(rule_id_to_completable_states_.size() - 1)) {
+      // It's a right recursion. We can optimize it.
+      // If it's the right recursion, we need to add the ancestors of the parent state.
+      if (state.rule_start_pos == ParserState::kNoPrevInputPos) {
+        // In this case, we can mark the new state as the root state to speed up.
+        right_recursion_to_root = true;
+      } else {
+        const auto in_vec = [&](const ParserState& state_) {
+          return std::find_if(
+                     rule_id_to_completable_states_.Back().begin(),
+                     rule_id_to_completable_states_.Back().end(),
+                     [&](const auto& s) {
+                       return StateEqualForParsing()(s.second, state_) && s.first == ref_rule_id;
+                     }
+                 ) != rule_id_to_completable_states_.Back().end();
+        };
+        const auto& parent_states_map = rule_id_to_completable_states_[state.rule_start_pos];
+        std::vector<std::pair<int32_t, ParserState>> to_added_states;
+        for (const auto& parent_state_iter : parent_states_map) {
+          if (parent_state_iter.first != state.rule_id) continue;
+          const auto& parent_state = parent_state_iter.second;
+          if (!in_vec(parent_state)) {
+            to_added_states.push_back({ref_rule_id, parent_state});
+          }
+        }
+        for (const auto& to_add_state : to_added_states) {
+          rule_id_to_completable_states_.PushBackInLatestRow(to_add_state);
+        }
+      }
+    } else {
+      // If it's not a right recursion, we need to add the current state.
+      rule_id_to_completable_states_.PushBackInLatestRow(
+          {ref_rule_id,
+           ParserState{state.rule_id, state.sequence_id, target, state.rule_start_pos, 0}}
+      );
+    }
+
+    // Check if the reference rule can be empty.
+    if (std::binary_search(
+            grammar_->allow_empty_rule_ids.begin(),
+            grammar_->allow_empty_rule_ids.end(),
+            ref_rule_id
+        )) {
+      Enqueue(ParserState{state.rule_id, state.sequence_id, target, state.rule_start_pos, 0});
+    }
+
+    // If the reference rule is not visited, we need to add it to the queue.
+    const auto& ref_rule = grammar_->GetRule(ref_rule_id);
+    const auto& ref_grammar_expr_id = ref_rule.body_expr_id;
+
+    if (grammar_->per_rule_fsms[ref_rule_id].has_value()) {
+      if (std::binary_search(
+              grammar_->allow_empty_rule_ids.begin(),
+              grammar_->allow_empty_rule_ids.end(),
+              ref_rule_id
+          )) {
+        Enqueue(ParserState{state.rule_id, state.sequence_id, target, state.rule_start_pos, 0});
+      }
+      const auto& ref_fsm = grammar_->per_rule_fsms[ref_rule_id].value();
+      Enqueue(ParserState{
+          ref_rule_id,
+          ref_grammar_expr_id,
+          ref_fsm.GetStart(),
+          right_recursion_to_root ? ParserState::kNoPrevInputPos
+                                  : int32_t(rule_id_to_completable_states_.size() - 1),
+          0
+      });
+    } else {
+      const auto& ref_grammar_expr = grammar_->GetGrammarExpr(ref_grammar_expr_id);
+      for (const auto& sequence_id : ref_grammar_expr) {
+        const auto& sequence = grammar_->GetGrammarExpr(sequence_id);
+        if (sequence.type == GrammarExprType::kEmptyStr) {
+          Enqueue(ParserState{state.rule_id, state.sequence_id, target, state.rule_start_pos, 0});
+          continue;
+        }
+        Enqueue(ParserState{
+            ref_rule_id,
+            sequence_id,
+            0,
+            right_recursion_to_root ? ParserState::kNoPrevInputPos
+                                    : int32_t(rule_id_to_completable_states_.size() - 1),
+            0
+        });
+      }
+    }
+  }
+}
+
 void EarleyParser::AdvanceByteString(
     const ParserState& state, const uint8_t ch, const GrammarExpr& sub_rule
 ) {
@@ -532,39 +708,20 @@ void EarleyParser::AdvanceCharacterClassStar(
   }
 }
 
-void EarleyParser::AdvanceTagDispatch(
-    const ParserState& state, const uint8_t ch, const GrammarExpr& cur_sequence
-) {
-  auto root_tag_dispatch_fsm_optional = grammar_->root_tag_dispatch_fsm;
-  if (!root_tag_dispatch_fsm_optional) {
-    XGRAMMAR_LOG(FATAL) << "The grammar does not have a root tag dispatch rule; it is not built.";
-    XGRAMMAR_UNREACHABLE();
-  }
-  auto root_tag_dispatch_fsm = root_tag_dispatch_fsm_optional.value();
-  auto start_node = root_tag_dispatch_fsm.GetStart();
-  auto next_node = root_tag_dispatch_fsm->GetNextState(state.element_id, ch);
-  auto new_state = state;
-  if (next_node == CompactFSM::kNoNextState) {
-    // Case 1. The new char cannot continue to be accepted by the tag dispatch fsm.
-    // We try to accept the new char from the start node. If accepted, we go to the target
-    // node. If it still cannot be accepted, we stay at the start node.
-    auto new_next_node = root_tag_dispatch_fsm->GetNextState(start_node, ch);
-    new_state.element_id = new_next_node == CompactFSM::kNoNextState ? start_node : new_next_node;
-    if (root_tag_dispatch_fsm.IsEndState(new_state.element_id)) {
-      tmp_process_state_queue_.push(new_state);
-    } else {
-      tmp_accept_stop_token_ = true;
-      tmp_states_to_be_added_.push_back(new_state);
+void EarleyParser::AdvanceFsm(const ParserState& state, const uint8_t ch) {
+  XGRAMMAR_DCHECK(state.rule_id != -1 && grammar_->per_rule_fsms[state.rule_id].has_value());
+  const auto& current_fsm = grammar_->per_rule_fsms[state.rule_id].value();
+  for (const auto& edge : current_fsm.GetFsm().GetEdges(state.element_id)) {
+    if ((!edge.IsCharRange()) || ch < edge.min || ch > edge.max) {
+      continue;
     }
-  } else {
-    // Case 2. The new char can continue to be accepted by the tag dispatch fsm.
-    // We need to update the element id to the next node.
-    new_state.element_id = next_node;
-    if (root_tag_dispatch_fsm.IsEndState(next_node)) {
-      tmp_process_state_queue_.push(new_state);
+    auto new_state = state;
+    new_state.element_id = edge.target;
+    if ((!current_fsm.IsNonTerminalState(edge.target)) &&
+        (!current_fsm.IsEndState(edge.target) && current_fsm.IsScanableState(edge.target))) {
+      EnqueueWithoutProcessing(std::move(new_state));
     } else {
-      tmp_accept_stop_token_ = true;
-      tmp_states_to_be_added_.push_back(new_state);
+      Enqueue(std::move(new_state));
     }
   }
 }
