@@ -16,13 +16,14 @@
 #include <cstdint>
 #include <functional>
 #include <string>
-#include <unordered_map>
 #include <unordered_set>
 #include <vector>
 
 #include "support/compact_2d_array.h"
-#include "support/reflection/reflection.h"
+#include "support/logging.h"
+#include "support/reflection.h"
 #include "support/utils.h"
+#include "xgrammar/exception.h"
 
 namespace xgrammar {
 
@@ -57,14 +58,14 @@ struct alignas(8) FSMEdge {
    */
   int32_t target;
 
+  // for serialization only
+  FSMEdge() = default;
+
   FSMEdge(int16_t min, int16_t max, int32_t target) : min(min), max(max), target(target) {
     XGRAMMAR_DCHECK(!IsCharRange() || min <= max)
         << "Invalid FSMEdge: min > max. min=" << min << ", max=" << max;
   }
 
-  // for serialization only
-  FSMEdge() = default;
-
   /*!
    * \brief Compare the edges. Used to sort the edges in the FSM.
    */
@@ -107,6 +108,8 @@ struct alignas(8) FSMEdge {
    * \return The rule id of the edge. -1 if the edge is not a rule reference.
    */
   int32_t GetRefRuleId() const { return IsRuleRef() ? max : -1; }
+
+  friend struct member_trait<FSMEdge>;
 };
 
 /*!
@@ -122,21 +125,9 @@ XGRAMMAR_MEMBER_ARRAY(FSMEdge, &FSMEdge::min, &FSMEdge::max, &FSMEdge::target);
 
 }  // namespace xgrammar
 
-namespace std {
-
-/*!
- * \brief Hash function for FSMEdge.
- */
-template <>
-struct hash<xgrammar::FSMEdge> {
-  size_t operator()(const xgrammar::FSMEdge& edge) const {
-    return std::hash<std::tuple<int16_t, int16_t, int32_t>>()(
-        std::make_tuple(edge.min, edge.max, edge.target)
-    );
-  }
-};
-
-}  // namespace std
+XGRAMMAR_HASH_BY_MEMBERS(
+    xgrammar::FSMEdge, &xgrammar::FSMEdge::min, &xgrammar::FSMEdge::max, &xgrammar::FSMEdge::target
+);
 
 namespace xgrammar {
 
@@ -299,9 +290,10 @@ class FSM {
    * \brief Add a whole FSM to the current FSM.
    * \param fsm The FSM to be added.
    * \param state_mapping The mapping from the state ids of the added FSM to the new ids in the
-   * current FSM. The result is cleared at the beginning.
+   * current FSM. The result is cleared at the beginning. If the fsm's state id starts from 0, use
+   * it for efficiency.
    */
-  void AddFSM(const FSM& fsm, std::unordered_map<int, int>* state_mapping = nullptr);
+  void AddFSM(const FSM& fsm, std::vector<int>* state_mapping = nullptr);
 
   /****************** FSM Construction Methods ******************/
 
@@ -316,7 +308,7 @@ class FSM {
    * \param new_num_states The new number of states.
    * \return The rebuilt FSM.
    */
-  FSM RebuildWithMapping(std::unordered_map<int, int>& state_mapping, int new_num_states) const;
+  FSM RebuildWithMapping(const std::vector<int>& state_mapping, int new_num_states) const;
 
   /*!
    * \brief Sort the edges of the FSM by their min, max and target.
@@ -343,14 +335,15 @@ class FSM {
  *
  * It share the same set of visitor methods with FSM.
  */
+
 class CompactFSM {
  public:
   // for serialization only
   CompactFSM() = default;
 
-  CompactFSM(const Compact2DArray<FSMEdge>& edges);
+  explicit CompactFSM(const Compact2DArray<FSMEdge>& edges);
 
-  CompactFSM(Compact2DArray<FSMEdge>&& edges);
+  explicit CompactFSM(Compact2DArray<FSMEdge>&& edges);
 
   /****************** CompactFSM Visitors ******************/
 
@@ -395,10 +388,15 @@ class CompactFSM {
    * transitions, the first one will be returned.
    * \param from The source state to transition from.
    * \param character The input character.
+   * \param targets The target states to be filled with the possible next states.
    * \return The target state if a valid transition exists, kNoNextState otherwise.
    */
-  int GetNextState(int from, int value, FSMEdge::EdgeType edge_type = FSMEdge::EdgeType::kCharRange)
-      const;
+  void GetNextStates(
+      int from,
+      int value,
+      FSMEdge::EdgeType edge_type = FSMEdge::EdgeType::kCharRange,
+      std::vector<int>* targets = nullptr
+  ) const;
 
   /*!
    * \brief Advance the FSM to the next state.
@@ -445,12 +443,18 @@ class CompactFSM {
    */
   FSM ToFSM() const;
 
-  picojson::value SerializeJSONValue() const;
-  friend void DeserializeJSONValue(CompactFSM& fsm, const picojson::value& v);
+  friend picojson::value SerializeJSONValue(const CompactFSM& value);
+  friend std::optional<SerializationError> DeserializeJSONValue(
+      CompactFSM* result, const picojson::value& value, const std::string& type_name
+  );
 
   XGRAMMAR_DEFINE_PIMPL_METHODS(CompactFSM);
 };
 
+std::optional<SerializationError> DeserializeJSONValue(
+    CompactFSM* result, const picojson::value& value, const std::string& type_name = ""
+);
+
 class CompactFSMWithStartEnd;
 
 /*!
@@ -465,35 +469,57 @@ class FSMWithStartEndBase {
   );
 
  public:
-  // for serialization only
+  // For serialization only
   FSMWithStartEndBase() = default;
 
-  /*! \brief Constructs an FSMWithStartEnd with a given FSM, start state, and end states. */
   FSMWithStartEndBase(
-      const FSMType& fsm, int start, const std::unordered_set<int>& ends, bool is_dfa = false
+      const FSMType& fsm, int start, const std::vector<bool>& ends, bool is_dfa = false
   )
       : fsm_(fsm), start_(start), ends_(ends), is_dfa_(is_dfa) {}
 
   /****************** Member Accessors and Mutators ******************/
 
   /*! \brief Returns the underlying FSM. */
-  const FSMType& GetFSM() const { return fsm_; }
+  const FSMType& GetFsm() const { return fsm_; }
 
   /*! \brief Returns the start state of the FSM. */
   int GetStart() const { return start_; }
 
   /*! \brief Returns the end states of the FSM. */
-  const std::unordered_set<int>& GetEnds() const { return ends_; }
+  const std::vector<bool>& GetEnds() const { return ends_; }
 
   /*!
    * \brief Checks if a given state is an end/accepting state.
    * \param state The state to check.
    * \return True if the state is an end state, false otherwise.
    */
-  bool IsEndState(int state) const {
-    return std::any_of(ends_.begin(), ends_.end(), [state](int end_state) {
-      return end_state == state;
-    });
+  bool IsEndState(int state) const { return ends_[state]; }
+
+  /*! \brief Check if a state is scanable.
+   *  \param state The state to check.
+   *  \return True if the state is scanable, false otherwise.
+   */
+  bool IsScanableState(int state) const {
+    for (const auto& edge : fsm_.GetEdges(state)) {
+      if (edge.IsCharRange()) {
+        return true;
+      }
+    }
+    return false;
+  }
+
+  /*!
+   * \brief Check if a state is not terminal.
+   * \param state The state to check.
+   * \return True if the state is scanable, false otherwise.
+   */
+  bool IsNonTerminalState(int state) const {
+    for (const auto& edge : fsm_.GetEdges(state)) {
+      if (edge.IsRuleRef() || edge.IsEpsilon()) {
+        return true;
+      }
+    }
+    return false;
   }
 
   /*!
@@ -511,14 +537,23 @@ class FSMWithStartEndBase {
    */
   void AddEndState(int state) {
     XGRAMMAR_DCHECK(state < NumStates());
-    ends_.insert(state);
+    ends_[state] = true;
+  }
+
+  /*!
+   * \brief Adds a new state to the FSM and marks it as non-end.
+   * \return The index of the newly added state.
+   */
+  int AddState() {
+    ends_.push_back(false);
+    return fsm_.AddState();
   }
 
   /*!
    * \brief Sets the end states of the FSM.
    * \param ends The new end states.
    */
-  void SetEndStates(const std::unordered_set<int>& ends) { ends_ = ends; }
+  void SetEndStates(const std::vector<bool>& ends) { ends_ = ends; }
 
   /*! \brief Returns the total number of states in the FSM. */
   int NumStates() const { return fsm_.NumStates(); }
@@ -526,12 +561,7 @@ class FSMWithStartEndBase {
   /*!
    * \brief Access the methods of the underlying FSM.
    */
-  FSMType* operator->() { return &fsm_; }
-
-  /*!
-   * \brief Access the methods of the underlying FSM.
-   */
-  const FSMType* operator->() const { return &fsm_; }
+  FSMType& GetFsm() { return fsm_; }
 
   /****************** FSM Traversal Algorithms ******************/
 
@@ -559,12 +589,13 @@ class FSMWithStartEndBase {
   FSMType fsm_;
   /*! \brief The start state of the FSM. */
   int start_;
+
   /*! \brief The set of accepting/end states. */
-  std::unordered_set<int> ends_;
+  std::vector<bool> ends_;
+
+ protected:
   /*! \brief Whether this FSM is a deterministic finite automaton. */
   bool is_dfa_ = false;
-
-  friend struct member_trait<CompactFSMWithStartEnd>;
 };
 
 /*!
@@ -597,9 +628,8 @@ class FSMWithStartEnd : public FSMWithStartEndBase<FSM> {
    * \param state_mapping The mapping from old state ids to new state ids.
    * \param new_num_states The new number of states.
    */
-  FSMWithStartEnd RebuildWithMapping(
-      std::unordered_map<int, int>& state_mapping, int new_num_states
-  );
+  FSMWithStartEnd RebuildWithMapping(const std::vector<int>& state_mapping, int new_num_states)
+      const;
 
   /*!
    * \brief Add the underlying FSM to another complete FSM that could contain multiple FSMs.
@@ -610,7 +640,7 @@ class FSMWithStartEnd : public FSMWithStartEndBase<FSM> {
    * cleared at the beginning. Should not be nullptr.
    * \return The FSMWithStartEnd that points to the complete FSM.
    */
-  FSMWithStartEnd AddToCompleteFSM(FSM* complete_fsm, std::unordered_map<int, int>* state_mapping);
+  FSMWithStartEnd AddToCompleteFSM(FSM* complete_fsm, std::vector<int>* state_mapping);
 
   /*!
    * \brief Transform the FSMWithStartEnd to a CompactFSMWithStartEnd.
@@ -642,7 +672,7 @@ class FSMWithStartEnd : public FSMWithStartEndBase<FSM> {
    * \brief Return a new FSM representing the complement of the language.
    * \return The complement FSM.
    */
-  FSMWithStartEnd Not() const;
+  Result<FSMWithStartEnd> Not(int max_result_num_states = 1e6) const;
 
   /*!
    * \brief Intersect the FSMs.
@@ -651,7 +681,7 @@ class FSMWithStartEnd : public FSMWithStartEndBase<FSM> {
    * \return The intersection of the FSMs.
    */
   static Result<FSMWithStartEnd> Intersect(
-      const FSMWithStartEnd& lhs, const FSMWithStartEnd& rhs, int num_of_states_limited = 1e6
+      const FSMWithStartEnd& lhs, const FSMWithStartEnd& rhs, int max_result_num_states = 1e6
   );
 
   /*!
@@ -679,7 +709,7 @@ class FSMWithStartEnd : public FSMWithStartEndBase<FSM> {
    * \details If a --\epsilon--> b, and either 1) b doesn't have any other inward edges, or
    * 2) a doesn't have any other outward edges, we can merge a and b.
    */
-  FSMWithStartEnd SimplifyEpsilon() const;
+  FSMWithStartEnd SimplifyEpsilon(int max_num_states = 1e8) const;
 
   /*!
    * \brief Merge equivalent states in the FSM.
@@ -687,19 +717,21 @@ class FSMWithStartEnd : public FSMWithStartEndBase<FSM> {
    * 2) they are not pointed to by other edges, then we can merge them.
    * \example n0 --(c)--> n1, n0 --(c)--> n2, then we can merge n1 and n2.
    */
-  FSMWithStartEnd MergeEquivalentSuccessors() const;
+  FSMWithStartEnd MergeEquivalentSuccessors(int max_num_states = 1e5) const;
 
   /*!
    * \brief Transform the FSM to a DFA.
+   * \param max_result_num_states The maximum number of states in the DFA.
    * \return The DFA.
    */
-  FSMWithStartEnd ToDFA() const;
+  Result<FSMWithStartEnd> ToDFA(int max_num_states = 1e3) const;
 
   /*!
    * \brief Minimize the DFA.
+   * \param max_result_num_states The maximum number of states in the DFA.
    * \return The minimized DFA.
    */
-  FSMWithStartEnd MinimizeDFA() const;
+  Result<FSMWithStartEnd> MinimizeDFA(int max_num_states = 1e3) const;
 };
 
 /*!
@@ -710,11 +742,11 @@ class FSMWithStartEnd : public FSMWithStartEndBase<FSM> {
  */
 class CompactFSMWithStartEnd : public FSMWithStartEndBase<CompactFSM> {
  public:
-  using FSMWithStartEndBase<CompactFSM>::FSMWithStartEndBase;
-
-  // for serialization only
+  // For serialization only
   CompactFSMWithStartEnd() = default;
 
+  using FSMWithStartEndBase<CompactFSM>::FSMWithStartEndBase;
+
   /*!
    * \brief Convert the FSMWithStartEnd to a string. Only considers the nodes approachable from the
    * start state.
@@ -722,6 +754,18 @@ class CompactFSMWithStartEnd : public FSMWithStartEndBase<CompactFSM> {
    */
   std::string ToString() const;
 
+  /*!
+   * \brief Transform the CompactFSMWithStartEnd to a FSMWithStartEnd.
+   * \return The FSMWithStartEnd.
+   */
+  FSMWithStartEnd ToFSM() const;
+
+  /*!
+   * \brief Print the CompactFSMWithStartEnd.
+   * \param os The output stream.
+   * \param fsm The CompactFSMWithStartEnd.
+   * \return The output stream.
+   */
   friend std::ostream& operator<<(std::ostream& os, const CompactFSMWithStartEnd& fsm);
 
   /*!
@@ -731,11 +775,14 @@ class CompactFSMWithStartEnd : public FSMWithStartEndBase<CompactFSM> {
    */
   friend std::size_t MemorySize(const CompactFSMWithStartEnd& self);
 
-  /*!
-   * \brief Transform the CompactFSMWithStartEnd to a FSMWithStartEnd.
-   * \return The FSMWithStartEnd.
-   */
-  FSMWithStartEnd ToFSM() const;
+  friend struct member_trait<CompactFSMWithStartEnd>;
+
+  friend struct CompactFSMWithStartEndSerializeHelper;
+
+  friend picojson::value SerializeJSONValue(const CompactFSMWithStartEnd& value);
+  friend std::optional<SerializationError> DeserializeJSONValue(
+      CompactFSMWithStartEnd* result, const picojson::value& value, const std::string& type_name
+  );
 };
 
 XGRAMMAR_MEMBER_ARRAY(
@@ -768,7 +815,7 @@ inline bool FSMWithStartEndBase<FSMType>::AcceptString(const std::string& str) c
     start_states = result_states;
   }
   return std::any_of(start_states.begin(), start_states.end(), [&](int state) {
-    return ends_.find(state) != ends_.end();
+    return ends_[state];
   });
 }
 
