@@ -4,11 +4,16 @@
  */
 #include "fsm_builder.h"
 
+#include <sys/types.h>
+
 #include <cstddef>
 #include <cstdint>
 #include <cstring>
+#include <optional>
+#include <set>
 #include <stack>
 #include <unordered_set>
+#include <utility>
 #include <variant>
 #include <vector>
 
@@ -187,7 +192,7 @@ Result<std::pair<int, int>> RegexIR::CheckRepeat(const std::string& regex, int&
 Result<FSMWithStartEnd> RegexIR::Build() const {
   if (states.empty()) {
     FSM empty_fsm(1);
-    FSMWithStartEnd result(empty_fsm, 0, std::unordered_set<int>{0}, false);
+    FSMWithStartEnd result(empty_fsm, 0, {true}, false);
     return ResultOk(std::move(result));
   }
   std::vector<FSMWithStartEnd> fsm_list;
@@ -283,7 +288,11 @@ Result<FSMWithStartEnd> RegexIR::visit(const RegexIR::Repeat& state) const {
 
   if (state.lower_bound == 1) {
     // Insert the first end state.
-    new_ends = result.GetEnds();
+    for (int end = 0; end < result.NumStates(); ++end) {
+      if (result.IsEndState(end)) {
+        new_ends.insert(end);
+      }
+    }
   }
 
   // Handling {n,}
@@ -291,10 +300,20 @@ Result<FSMWithStartEnd> RegexIR::visit(const RegexIR::Repeat& state) const {
     for (int i = 2; i < state.lower_bound; i++) {
       result = FSMWithStartEnd::Concat(std::vector<FSMWithStartEnd>{result, child});
     }
-    int end_state_of_lower_bound_fsm = *result.GetEnds().begin();
+    int end_state_of_lower_bound_fsm = -1;
+    for (int end = 0; end < result.NumStates(); ++end) {
+      if (result.IsEndState(end)) {
+        end_state_of_lower_bound_fsm = end;
+        break;
+      }
+    }
+    XGRAMMAR_DCHECK(end_state_of_lower_bound_fsm != -1)
+        << "No end state found in the lower bound FSM.";
     result = FSMWithStartEnd::Concat(std::vector<FSMWithStartEnd>{result, child});
-    for (const auto& end : result.GetEnds()) {
-      result->AddEpsilonEdge(end, end_state_of_lower_bound_fsm);
+    for (int end = 0; end < result.NumStates(); ++end) {
+      if (result.IsEndState(end)) {
+        result.GetFsm().AddEpsilonEdge(end, end_state_of_lower_bound_fsm);
+      }
     }
     return ResultOk(std::move(result));
   }
@@ -302,8 +321,10 @@ Result<FSMWithStartEnd> RegexIR::visit(const RegexIR::Repeat& state) const {
   for (int i = 2; i <= state.upper_bound; i++) {
     result = FSMWithStartEnd::Concat(std::vector<FSMWithStartEnd>{result, child});
     if (i >= state.lower_bound) {
-      for (const auto& end : result.GetEnds()) {
-        new_ends.insert(end);
+      for (int end = 0; end < result.NumStates(); ++end) {
+        if (result.IsEndState(end)) {
+          new_ends.insert(end);
+        }
       }
     }
   }
@@ -315,64 +336,70 @@ Result<FSMWithStartEnd> RegexIR::visit(const RegexIR::Repeat& state) const {
 
 FSMWithStartEnd RegexIR::BuildLeafFSMFromRegex(const std::string& regex) {
   FSM empty_fsm(0);
-  FSMWithStartEnd result(empty_fsm, 0, std::unordered_set<int>{}, true);
+  FSMWithStartEnd result(empty_fsm, 0, {}, true);
   // Handle the regex string.
   if (!(regex[0] == '[' && regex[regex.size() - 1] == ']')) {
-    result->AddState();
+    result.AddState();
     for (size_t i = 0; i < regex.size(); i++) {
       if (regex[i] != '\\') {
         if (regex[i] == '.') {
-          result->AddEdge(result->NumStates() - 1, result->NumStates(), 0, 0xFF);
+          result.GetFsm().AddEdge(result.NumStates() - 1, result.NumStates(), 0, 0xFF);
         } else {
-          result->AddEdge(
-              result->NumStates() - 1,
-              result->NumStates(),
+          result.GetFsm().AddEdge(
+              result.NumStates() - 1,
+              result.NumStates(),
               static_cast<uint8_t>(regex[i]),
               static_cast<uint8_t>(regex[i])
           );
         }
-        result->AddState();
+        result.AddState();
         continue;
       }
       std::vector<std::pair<int, int>> escape_vector = HandleEscapes(regex, i);
       for (const auto& escape : escape_vector) {
-        result->AddEdge(
-            result->NumStates() - 1,
-            result->NumStates(),
+        result.GetFsm().AddEdge(
+            result.NumStates() - 1,
+            result.NumStates(),
             static_cast<uint8_t>(escape.first),
             static_cast<uint8_t>(escape.second)
         );
       }
-      result->AddState();
+      result.AddState();
       i++;
     }
-    result.AddEndState(result->NumStates() - 1);
+    result.AddEndState(result.NumStates() - 1);
   } else if (regex[0] == '[' && regex[regex.size() - 1] == ']') {
     // Handle the character class.
-    result->AddState();
-    result->AddState();
+    result.AddState();
+    result.AddState();
     result.AddEndState(1);
     bool reverse = regex[1] == '^';
     for (size_t i = reverse ? 2 : 1; i < regex.size() - 1; i++) {
       if (regex[i] != '\\') {
         if (!(((i + 2) < regex.size() - 1) && regex[i + 1] == '-')) {
           // A single char.
-          result->AddEdge(0, 1, static_cast<uint8_t>(regex[i]), static_cast<uint8_t>(regex[i]));
+          result.GetFsm().AddEdge(
+              0, 1, static_cast<uint8_t>(regex[i]), static_cast<uint8_t>(regex[i])
+          );
           continue;
         }
         // Handle the char range.
         if (regex[i + 2] != '\\') {
-          result->AddEdge(0, 1, static_cast<uint8_t>(regex[i]), static_cast<uint8_t>(regex[i + 2]));
+          result.GetFsm().AddEdge(
+              0, 1, static_cast<uint8_t>(regex[i]), static_cast<uint8_t>(regex[i + 2])
+          );
           i = i + 2;
           continue;
         }
         auto escaped_edges = HandleEscapes(regex, i + 2);
         // Means it's not a range.
         if (escaped_edges.size() != 1 || escaped_edges[0].first != escaped_edges[0].second) {
-          result->AddEdge(0, 1, static_cast<uint8_t>(regex[i]), static_cast<uint8_t>(regex[i]));
+          result.GetFsm().AddEdge(
+              0, 1, static_cast<uint8_t>(regex[i]), static_cast<uint8_t>(regex[i])
+          );
           continue;
         }
-        result->AddEdge(
+        result.GetFsm().AddEdge(
             0, 1, static_cast<uint8_t>(regex[0]), static_cast<uint8_t>(escaped_edges[0].first)
         );
         i = i + 3;
@@ -383,14 +410,14 @@ FSMWithStartEnd RegexIR::BuildLeafFSMFromRegex(const std::string& regex) {
       if (escaped_edges.size() != 1 || escaped_edges[0].first != escaped_edges[0].second) {
         // It's a multi-match escape char.
         for (const auto& edge : escaped_edges) {
-          result->AddEdge(
+          result.GetFsm().AddEdge(
               0, 1, static_cast<uint8_t>(edge.first), static_cast<uint8_t>(edge.second)
           );
         }
         continue;
       }
       if (!(((i + 2) < regex.size() - 1) && regex[i + 1] == '-')) {
-        result->AddEdge(
+        result.GetFsm().AddEdge(
             0,
             1,
             static_cast<uint8_t>(escaped_edges[0].first),
@@ -399,7 +426,7 @@ FSMWithStartEnd RegexIR::BuildLeafFSMFromRegex(const std::string& regex) {
         continue;
       }
       if (regex[i + 2] != '\\') {
-        result->AddEdge(
+        result.GetFsm().AddEdge(
             0, 1, static_cast<uint8_t>(escaped_edges[0].first), static_cast<uint8_t>(regex[i + 2])
         );
         i = i + 2;
@@ -408,7 +435,7 @@ FSMWithStartEnd RegexIR::BuildLeafFSMFromRegex(const std::string& regex) {
       auto rhs_escaped_edges = HandleEscapes(regex, i + 2);
       if (rhs_escaped_edges.size() != 1 ||
           rhs_escaped_edges[0].first != rhs_escaped_edges[0].second) {
-        result->AddEdge(
+        result.GetFsm().AddEdge(
             0,
             1,
             static_cast<uint8_t>(escaped_edges[0].first),
@@ -416,7 +443,7 @@ FSMWithStartEnd RegexIR::BuildLeafFSMFromRegex(const std::string& regex) {
         );
         continue;
       }
-      result->AddEdge(
+      result.GetFsm().AddEdge(
           0,
           1,
           static_cast<uint8_t>(escaped_edges[0].first),
@@ -428,7 +455,7 @@ FSMWithStartEnd RegexIR::BuildLeafFSMFromRegex(const std::string& regex) {
     bool has_edge[0x100];
     memset(has_edge, 0, sizeof(has_edge));
     FSM new_fsm(2);
-    for (const auto& edge : result->GetEdges(0)) {
+    for (const auto& edge : result.GetFsm().GetEdges(0)) {
       for (int i = edge.min; i <= edge.max; i++) {
         has_edge[i] = true;
       }
@@ -468,7 +495,9 @@ FSMWithStartEnd RegexIR::BuildLeafFSMFromRegex(const std::string& regex) {
         new_fsm.AddEdge(0, 1, last, 0xFF);
       }
     }
-    result = FSMWithStartEnd(new_fsm, 0, std::unordered_set<int>{1}, false);
+    std::vector<bool> ends(new_fsm.NumStates(), false);
+    ends[1] = true;
+    result = FSMWithStartEnd(new_fsm, 0, ends, false);
   } else {
     // TODO: The support for rules.
     XGRAMMAR_LOG(WARNING) << "rule is not supported yet.";
@@ -747,8 +776,23 @@ Result<FSMWithStartEnd> RegexFSMBuilder::Build(const std::string& regex) {
   return ir.Build();
 }
 
-FSMWithStartEnd TrieFSMBuilder::Build(
-    const std::vector<std::string>& patterns, std::vector<int32_t>* end_states
+class TrieFSMBuilderImpl {
+ public:
+  TrieFSMBuilderImpl() = default;
+  std::optional<FSMWithStartEnd> Build(
+      const std::vector<std::string>& patterns,
+      std::vector<int32_t>* end_states,
+      bool allow_overlap,
+      bool add_back_edges
+  );
+  void AddBackEdges(FSM* fsm, int start, const std::unordered_set<int>& ends);
+};
+
+std::optional<FSMWithStartEnd> TrieFSMBuilderImpl::Build(
+    const std::vector<std::string>& patterns,
+    std::vector<int32_t>* end_states,
+    bool allow_overlap,
+    bool add_back_edges
 ) {
   FSM fsm(1);
   int start = 0;
@@ -759,6 +803,11 @@ FSMWithStartEnd TrieFSMBuilder::Build(
   }
 
   for (const auto& pattern : patterns) {
+    // Check for empty patterns
+    if (!allow_overlap && pattern.empty()) {
+      return std::nullopt;
+    }
+
     int current_state = 0;
     for (const auto& ch : pattern) {
       int16_t ch_int16 = static_cast<int16_t>(static_cast<uint8_t>(ch));
@@ -768,13 +817,101 @@ FSMWithStartEnd TrieFSMBuilder::Build(
         fsm.AddEdge(current_state, next_state, ch_int16, ch_int16);
       }
       current_state = next_state;
+      if (!allow_overlap && ends.count(current_state) > 0) {
+        return std::nullopt;
+      }
+    }
+    if (!allow_overlap && fsm.GetEdges(current_state).size() > 0) {
+      return std::nullopt;
     }
     ends.insert(current_state);
     if (end_states) {
       end_states->push_back(current_state);
     }
   }
-  return FSMWithStartEnd(fsm, start, ends, true);
+  if (add_back_edges) {
+    AddBackEdges(&fsm, start, ends);
+  }
+
+  std::vector<bool> is_end_state(fsm.NumStates(), false);
+  for (const auto& end : ends) {
+    is_end_state[end] = true;
+  }
+
+  return FSMWithStartEnd(fsm, start, is_end_state);
+}
+
+void TrieFSMBuilderImpl::AddBackEdges(FSM* fsm, int start, const std::unordered_set<int>& ends) {
+  // Build an Aho-Corasick automaton by adding back edges.
+  // When matching on the trie fails, we should go back to the start state and
+  // find the next match. Back edges represent such state transitions.
+
+  auto f_add_range_edges = [&](int node, std::set<FSMEdge, FSMEdgeRangeComparator>& cur_edges_set) {
+    cur_edges_set.insert(FSMEdge(-1, -1, 0));
+    cur_edges_set.insert(FSMEdge(256, 256, 0));
+    XGRAMMAR_DCHECK(cur_edges_set.size() >= 2);
+    for (auto it = std::next(cur_edges_set.begin()); it != cur_edges_set.end(); ++it) {
+      FSMEdge prev_edge = *std::prev(it);
+      XGRAMMAR_DCHECK(prev_edge.max < it->min);
+      if (prev_edge.max + 1 != it->min) {
+        auto new_edge = FSMEdge(prev_edge.max + 1, it->min - 1, start);
+        // The new edge should be inserted before the current edge to avoid infinite loop.
+        XGRAMMAR_DCHECK(new_edge < *it);
+        cur_edges_set.insert(new_edge);
+      }
+    }
+
+    // Remove first and last element of cur_edges_set
+    XGRAMMAR_DCHECK(*cur_edges_set.begin() == FSMEdge(-1, -1, 0));
+    XGRAMMAR_DCHECK(*std::prev(cur_edges_set.end()) == FSMEdge(256, 256, 0));
+    cur_edges_set.erase(cur_edges_set.begin());
+    cur_edges_set.erase(std::prev(cur_edges_set.end()));
+
+    XGRAMMAR_DCHECK(cur_edges_set.begin()->min == 0);
+    XGRAMMAR_DCHECK(std::prev(cur_edges_set.end())->max == 255);
+  };
+
+  for (int i = 0; i < fsm->NumStates(); i++) {
+    if (i == start || ends.count(i) > 0) {
+      continue;
+    }
+    std::vector<FSMEdge>& cur_edges = fsm->GetEdges(i);
+    XGRAMMAR_DCHECK(cur_edges.size() > 0);
+    std::set<FSMEdge, FSMEdgeRangeComparator> cur_edges_set(cur_edges.begin(), cur_edges.end());
+
+    // Step 1. Add edges in the edges of the start state.
+    // For start--(c)-->t, add i--(c)-->t.
+    const auto& root_edges = fsm->GetEdges(start);
+    for (const auto& root_edge : root_edges) {
+      XGRAMMAR_DCHECK(root_edge.min == root_edge.max);
+      if (cur_edges_set.count(root_edge) == 0) {
+        cur_edges_set.insert(root_edge);
+      }
+    }
+
+    // Step 2. Add i--(c)-->start for c not in the edge set of i.
+    f_add_range_edges(i, cur_edges_set);
+
+    // Step 3. Update the edges of i.
+    cur_edges.clear();
+    cur_edges.insert(cur_edges.end(), cur_edges_set.begin(), cur_edges_set.end());
+  }
+
+  // Finally, add range edges to the start state.
+  std::vector<FSMEdge>& start_edges = fsm->GetEdges(start);
+  std::set<FSMEdge, FSMEdgeRangeComparator> start_edges_set(start_edges.begin(), start_edges.end());
+  f_add_range_edges(start, start_edges_set);
+  start_edges.clear();
+  start_edges.insert(start_edges.end(), start_edges_set.begin(), start_edges_set.end());
+}
+
+std::optional<FSMWithStartEnd> TrieFSMBuilder::Build(
+    const std::vector<std::string>& patterns,
+    std::vector<int32_t>* end_states,
+    bool allow_overlap,
+    bool add_back_edges
+) {
+  return TrieFSMBuilderImpl().Build(patterns, end_states, allow_overlap, add_back_edges);
 }
 
 }  // namespace xgrammar
