@@ -19,14 +19,12 @@ namespace xgrammar {
  */
 class RegexFSMBuilder {
  public:
-  RegexFSMBuilder() = default;
-
   /*!
    * \brief Converts a regex string to a FSM.
    * \param regex The regex string.
    * \return The FSM with start and end states.
    */
-  Result<FSMWithStartEnd> Build(const std::string& regex);
+  static Result<FSMWithStartEnd> Build(const std::string& regex);
 };
 
 /*!
@@ -34,17 +32,23 @@ class RegexFSMBuilder {
  */
 class TrieFSMBuilder {
  public:
-  TrieFSMBuilder() = default;
-
   /*!
    * \brief Build a trie-based FSM from a list of patterns.
    * \param patterns The patterns to be built.
    * \param end_states The end states of the FSM. This is the terminal state of each pattern and
    * the order follows the order of patterns.
-   * \return The FSM with start and end states.
+   * \param allow_overlap Whether to allow overlap between patterns (one being a prefix of the
+   * other). It does not allow empty patterns either. If false and there is overlap, will return
+   * std::nullopt.
+   * \param add_back_edges Whether to add back edges to the FSM. This complements the trie to an
+   * Aho-Corasick automaton.
+   * \return If success, the FSM with start and end states. Otherwise, std::nullopt.
    */
-  FSMWithStartEnd Build(
-      const std::vector<std::string>& patterns, std::vector<int32_t>* end_states = nullptr
+  static std::optional<FSMWithStartEnd> Build(
+      const std::vector<std::string>& patterns,
+      std::vector<int32_t>* end_states = nullptr,
+      bool allow_overlap = true,
+      bool add_back_edges = false
   );
 };
 
