@@ -12,7 +12,7 @@
 #include <cstdint>
 #include <unordered_map>
 
-#include "grammar_data_structure.h"
+#include "grammar_impl.h"
 
 namespace xgrammar {
 
@@ -95,7 +95,10 @@ class GrammarBuilder {
    * \param str The string to be added.
    */
   int32_t AddByteString(const std::string& str) {
-    std::vector<int32_t> bytes(str.begin(), str.end());
+    std::vector<int32_t> bytes;
+    for (char c : str) {
+      bytes.push_back(static_cast<int32_t>(static_cast<uint8_t>(c)));
+    }
     return AddGrammarExpr(
         {GrammarExprType::kByteString, bytes.data(), static_cast<int32_t>(bytes.size())}
     );
@@ -180,19 +183,34 @@ class GrammarBuilder {
    * \brief Add a GrammarExpr for tag dispatch.
    * \param tag_dispatch_list A list of pairs of tag_expr_id and rule_id.
    */
-  int32_t AddTagDispatch(const std::vector<std::pair<int32_t, int32_t>>& tag_dispatch_list) {
+  int32_t AddTagDispatch(const Grammar::Impl::TagDispatch& tag_dispatch) {
     std::vector<int32_t> data;
-    data.reserve(tag_dispatch_list.size() * 2);
-    for (const auto& [tag_expr_id, rule_id] : tag_dispatch_list) {
-      data.push_back(tag_expr_id);
+    data.reserve(tag_dispatch.tag_rule_pairs.size() * 2 + 3);
+    for (const auto& [tag, rule_id] : tag_dispatch.tag_rule_pairs) {
+      data.push_back(AddByteString(tag));
       data.push_back(rule_id);
     }
+    data.push_back(static_cast<int32_t>(tag_dispatch.stop_eos));
+    std::vector<int32_t> stop_str_expr_ids;
+    for (const auto& stop_str : tag_dispatch.stop_str) {
+      stop_str_expr_ids.push_back(AddByteString(stop_str));
+    }
+    data.push_back(AddChoices(stop_str_expr_ids));
+    data.push_back(static_cast<int32_t>(tag_dispatch.loop_after_dispatch));
     return AddGrammarExpr(
         {GrammarExprType::kTagDispatch, data.data(), static_cast<int32_t>(data.size())}
     );
   }
 
+  int32_t AddRepeat(int32_t ref_rule_id, int32_t min_repeat_count, int32_t max_repeat_count) {
+    std::vector<int32_t> data({ref_rule_id, min_repeat_count, max_repeat_count});
+    return AddGrammarExpr({GrammarExprType::kRepeat, data.data(), static_cast<int32_t>(data.size())}
+    );
+  }
+
+  /*! \brief Get the number of grammar_exprs. */
   int32_t NumGrammarExprs() const { return grammar_->NumGrammarExprs(); }
+
   /*! \brief Get the grammar_expr with the given id. */
   GrammarExpr GetGrammarExpr(int32_t grammar_expr_id) {
     return grammar_->GetGrammarExpr(grammar_expr_id);
@@ -203,7 +221,6 @@ class GrammarBuilder {
   /*! \brief Add a rule and return the rule id. */
   int32_t AddRule(const Rule& rule) {
     int32_t id = grammar_->rules_.size();
-    auto rules = grammar_->rules_;
     grammar_->rules_.push_back(rule);
     XGRAMMAR_CHECK(rule_name_to_id_.count(rule.name) == 0);
     rule_name_to_id_[rule.name] = id;
@@ -231,6 +248,10 @@ class GrammarBuilder {
    */
   int32_t AddEmptyRule(const std::string& name) { return AddRule({name, -1}); }
 
+  int32_t AddEmptyRuleWithHint(const std::string& name_hint) {
+    return AddRule({GetNewRuleName(name_hint), -1});
+  }
+
   /*!
    * \brief Update the rule body of the given rule, specified by rule id. Can be used to set the
    * rule body of a rule inserted by GrammarBuilder::AddEmptyRule.
@@ -261,6 +282,12 @@ class GrammarBuilder {
     grammar_->rules_[rule_id].lookahead_assertion_id = lookahead_assertion_id;
   }
 
+  void UpdateLookaheadExact(int32_t rule_id, bool is_exact = true) {
+    XGRAMMAR_CHECK(rule_id < static_cast<int32_t>(grammar_->rules_.size()))
+        << "Rule id " << rule_id << " is out of range.";
+    grammar_->rules_[rule_id].is_exact_lookahead = is_exact;
+  }
+
   /*!
    * \brief Add a lookahead assertion to a rule referred by the given name. The lookahead
    * assertion should be a sequence GrammarExpr id. An id of -1 means no lookahead assertion.
