@@ -12,7 +12,8 @@
 #include <string>
 
 #include "grammar_builder.h"
-#include "grammar_data_structure.h"
+#include "grammar_impl.h"
+#include "xgrammar/grammar.h"
 
 namespace xgrammar {
 
@@ -39,7 +40,7 @@ class GrammarFunctor {
    */
   virtual ReturnType Apply(const Grammar& grammar) {
     // The initializer MUST be called at first when overriding the Apply() function.
-    Init(grammar);
+    InitGrammar(grammar);
     if constexpr (std::is_same<T, void>::value) {
       for (int i = 0; i < static_cast<int>(base_grammar_->NumRules()); ++i) {
         auto rule = base_grammar_->GetRule(i);
@@ -50,20 +51,21 @@ class GrammarFunctor {
       return ReturnType();
     } else if constexpr (std::is_same<T, int32_t>::value &&
                          std::is_same<ReturnType, Grammar>::value) {
+      InitBuilder();
       // First add empty rules to ensure the new rule ids the same as the old ones, then update
       // the rule bodies
       for (int i = 0; i < static_cast<int>(base_grammar_->NumRules()); ++i) {
-        builder_.AddEmptyRule(base_grammar_->GetRule(i).name);
+        builder_->AddEmptyRule(base_grammar_->GetRule(i).name);
       }
       for (int i = 0; i < static_cast<int>(base_grammar_->NumRules()); ++i) {
         auto rule = base_grammar_->GetRule(i);
         cur_rule_name_ = rule.name;
         auto new_body_expr_id = VisitExpr(rule.body_expr_id);
-        builder_.UpdateRuleBody(i, new_body_expr_id);
+        builder_->UpdateRuleBody(i, new_body_expr_id);
         // Handle lookahead assertion
-        builder_.UpdateLookaheadAssertion(i, VisitLookaheadAssertion(rule.lookahead_assertion_id));
+        builder_->UpdateLookaheadAssertion(i, VisitLookaheadAssertion(rule.lookahead_assertion_id));
       }
-      return builder_.Get(base_grammar_->GetRootRule().name);
+      return builder_->Get(base_grammar_->GetRootRule().name);
     } else {
       return ReturnType();
     }
@@ -78,16 +80,22 @@ class GrammarFunctor {
   using GrammarExprType = Grammar::Impl::GrammarExprType;
 
   /*! \brief Initialize the functor. Should be called at the beginning of Apply(). */
-  virtual void Init(const Grammar& grammar) {
-    base_grammar_ = grammar;
-    builder_ = GrammarBuilder();
+  virtual void InitGrammar() {}
+
+  virtual void InitGrammar(const Grammar& grammar) { base_grammar_ = grammar; }
+
+  virtual void InitBuilder() {
+    owned_builder_ = GrammarBuilder();
+    builder_ = &owned_builder_;
   }
 
-  virtual void InitWithCopy(const Grammar& grammar) {
-    base_grammar_ = grammar;
-    builder_ = GrammarBuilder(grammar);
+  virtual void InitBuilder(const Grammar& grammar) {
+    owned_builder_ = GrammarBuilder(grammar);
+    builder_ = &owned_builder_;
   }
 
+  virtual void InitBuilder(GrammarBuilder* builder) { builder_ = builder; }
+
   /*! \brief Visit a lookahead assertion expr referred by id. */
   virtual T VisitLookaheadAssertion(int32_t lookahead_assertion_id) {
     if (lookahead_assertion_id == -1) {
@@ -124,8 +132,11 @@ class GrammarFunctor {
         return VisitRuleRef(grammar_expr);
       case GrammarExprType::kTagDispatch:
         return VisitTagDispatch(grammar_expr);
+      case GrammarExprType::kRepeat:
+        return VisitRepeat(grammar_expr);
       default:
         XGRAMMAR_LOG(FATAL) << "Unexpected sequence type: " << static_cast<int>(grammar_expr.type);
+        XGRAMMAR_UNREACHABLE();
     }
   }
 
@@ -140,7 +151,7 @@ class GrammarFunctor {
       for (int32_t i : grammar_expr) {
         choice_ids.push_back(VisitExpr(i));
       }
-      return builder_.AddChoices(choice_ids);
+      return builder_->AddChoices(choice_ids);
     } else {
       return T();
     }
@@ -157,7 +168,7 @@ class GrammarFunctor {
       for (int32_t i : grammar_expr) {
         sequence_ids.push_back(VisitExpr(i));
       }
-      return builder_.AddSequence(sequence_ids);
+      return builder_->AddSequence(sequence_ids);
     } else {
       return T();
     }
@@ -165,15 +176,10 @@ class GrammarFunctor {
 
   virtual T VisitTagDispatch(const GrammarExpr& grammar_expr) {
     if constexpr (std::is_same<T, void>::value) {
-      for (int i = 0; i < grammar_expr.size(); i += 2) {
-        VisitExpr(grammar_expr[i]);
-      }
+      return;
     } else if constexpr (std::is_same<T, int32_t>::value) {
-      std::vector<std::pair<int32_t, int32_t>> tag_dispatch_list;
-      for (int i = 0; i < grammar_expr.size(); i += 2) {
-        tag_dispatch_list.push_back({VisitExpr(grammar_expr[i]), grammar_expr[i + 1]});
-      }
-      return builder_.AddTagDispatch(tag_dispatch_list);
+      Grammar::Impl::TagDispatch tag_dispatch = base_grammar_->GetTagDispatch(grammar_expr);
+      return builder_->AddTagDispatch(tag_dispatch);
     } else {
       return T();
     }
@@ -184,7 +190,7 @@ class GrammarFunctor {
     if constexpr (std::is_same<T, void>::value) {
       return;
     } else if constexpr (std::is_same<T, int32_t>::value) {
-      return builder_.AddGrammarExpr(grammar_expr);
+      return builder_->AddGrammarExpr(grammar_expr);
     } else {
       return T();
     }
@@ -209,6 +215,9 @@ class GrammarFunctor {
   /*! \brief Visit a rule reference GrammarExpr. */
   virtual T VisitRuleRef(const GrammarExpr& grammar_expr) { return VisitElement(grammar_expr); }
 
+  /*! \brief Visit a repeat GrammarExpr. */
+  virtual T VisitRepeat(const GrammarExpr& grammar_expr) { return VisitElement(grammar_expr); }
+
   /*! \brief The grammar to visit or mutate. */
   Grammar base_grammar_{NullObj{}};
 
@@ -216,7 +225,9 @@ class GrammarFunctor {
    * \brief The builder to build the new grammar. It is empty when the mutator is constructed, and
    * can be used to build a new grammar in subclasses.
    */
-  GrammarBuilder builder_;
+  GrammarBuilder* builder_ = nullptr;
+
+  GrammarBuilder owned_builder_;
 
   /*! \brief The name of the current rule being visited. */
   std::string cur_rule_name_;
@@ -234,18 +245,9 @@ using GrammarVisitor = GrammarFunctor<void, ReturnType>;
  */
 using GrammarMutator = GrammarFunctor<int32_t, Grammar>;
 
-/*************************** Grammar manipulation methods ***************************/
 /****** All below methods are implemented as functor to hide the implementation ******/
 
-/*!
- * \brief Normalize a Grammar: expand the nested rules, combine consequent sequences and strings,
- * etc.
- */
-class GrammarNormalizer {
- public:
-  static Grammar Apply(const Grammar& grammar);
-};
-
+/*************************** Grammar Constructor ***************************/
 /*!
  * \brief Find the union of multiple grammars as a new grammar.
  */
@@ -263,28 +265,23 @@ class GrammarConcatFunctor {
 };
 
 /*!
- * \brief Analyze the grammar to find the rules that are allowed to be empty.
+ * \brief Add a sub grammar to the current builder. The return value
+ * of Apply is the new rule id of the sub grammar's root rule.
  */
-class AllowEmptyRuleAnalyzer {
+class SubGrammarAdder {
  public:
-  static std::vector<int32_t> Apply(const Grammar& grammar);
+  static int32_t Apply(GrammarBuilder* builder, const Grammar& sub_grammar);
 };
 
+/*************************** Grammar Normalizer ***************************/
+
 /*!
- * \brief Create a grammar that recognizes structural tags based on their triggers. See
- * StructuralTagToGrammar() for more details.
- *
- * \param triggers The trigger strings that identify each tag group
- * \param tag_groups The tags and their schema grammars, grouped by trigger. tag_groups[i][j] is the
- * j-th tag that matches triggers[i], and its corresponding schema grammar.
- * \return A grammar that matches all the tagged patterns.
+ * \brief Normalize a Grammar: expand the nested rules, combine consequent sequences and strings,
+ * etc.
  */
-class StructuralTagGrammarCreator {
+class GrammarNormalizer {
  public:
-  static Grammar Apply(
-      const std::vector<std::string>& triggers,
-      const std::vector<std::vector<std::pair<StructuralTagItem, Grammar>>>& tag_groups
-  );
+  static Grammar Apply(const Grammar& grammar);
 };
 
 /*!
@@ -296,6 +293,8 @@ class StructureNormalizer {
   static Grammar Apply(const Grammar& grammar);
 };
 
+/*************************** Grammar Optimizer ***************************/
+
 /*!
  * \brief Fuse the byte string elements in the grammar.
  */
@@ -304,6 +303,14 @@ class ByteStringFuser {
   static Grammar Apply(const Grammar& grammar);
 };
 
+/*!
+ * \brief Analyze the grammar to find the rules that are allowed to be empty.
+ */
+class AllowEmptyRuleAnalyzer {
+ public:
+  static std::vector<int32_t> Apply(const Grammar& grammar);
+};
+
 /*!
  * \brief Inline the rule references in the grammar.
  */
@@ -328,6 +335,49 @@ class LookaheadAssertionAnalyzer {
   static Grammar Apply(const Grammar& grammar);
 };
 
+/*!
+ * \brief Build the FSMs of the grammar.
+ */
+class GrammarFSMBuilder {
+  using GrammarExpr = Grammar::Impl::GrammarExpr;
+
+ public:
+  static void Apply(Grammar* grammar);
+  static FSMWithStartEnd RuleRef(const GrammarExpr& expr);
+  static FSMWithStartEnd CharacterClass(const GrammarExpr& expr);
+  static FSMWithStartEnd ByteString(const GrammarExpr& expr);
+  static std::optional<FSMWithStartEnd> Sequence(const GrammarExpr& expr, const Grammar& grammar);
+  static std::optional<FSMWithStartEnd> Choices(const GrammarExpr& expr, const Grammar& grammar);
+  static std::optional<FSMWithStartEnd> TagDispatch(const Grammar::Impl::TagDispatch& tag_dispatch);
+};
+
+/*!
+ * \brief Normalize the repetition expression. If the context of
+ * repetition expression is nullable, then the repetition range will be
+ * normalized from {m, n} to {0, n} to reduce uncertainty.
+ */
+class RepetitionNormalizer {
+ public:
+  static void Apply(Grammar* grammar);
+};
+
+/*!
+ * \brief Optimize the grammar when compiling.
+ * \note No matter whether the grammar is optimized, grammar optimizer will
+ * return a new grammar. The following optimization will be applied:
+ * 1. Byte fuser.
+ * 2. Rule inliner.
+ * 3. Dead code eliminator.
+ * 4. Lookahead assertion analyzer.
+ * 5. Allow-empty rule analyzer.
+ * 6. Repetition normalizer.
+ * 7. FSM builder.
+ */
+class GrammarOptimizer {
+ public:
+  static Grammar Apply(const Grammar& grammar);
+};
+
 }  // namespace xgrammar
 
 #endif  // XGRAMMAR_GRAMMAR_FUNCTOR_H_
