@@ -8,17 +8,22 @@
 
 #include <xgrammar/matcher.h>
 
+#include <algorithm>
 #include <cstdint>
+#include <optional>
+#include <thread>
 #include <utility>
+#include <variant>
 #include <vector>
 
-#include "compiled_grammar_data_structure.h"
+#include "compiled_grammar_impl.h"
 #include "earley_parser.h"
-#include "grammar_data_structure.h"
+#include "grammar_impl.h"
 #include "support/dynamic_bitset.h"
 #include "support/encoding.h"
 #include "support/int_set.h"
 #include "support/logging.h"
+#include "support/thread_pool.h"
 #include "testing.h"
 
 namespace xgrammar {
@@ -113,10 +118,12 @@ void ApplyTokenBitmaskInplaceCPU(
       logits->ndim == 2
           ? std::make_pair(static_cast<int>(logits->shape[0]), static_cast<int>(logits->shape[1]))
           : std::make_pair(1, static_cast<int>(logits->shape[0]));
+  int logits_stride0 = logits->strides[0];
   std::pair<int, int> bitmask_shape =
       bitmask.ndim == 2
           ? std::make_pair(static_cast<int>(bitmask.shape[0]), static_cast<int>(bitmask.shape[1]))
           : std::make_pair(1, static_cast<int>(bitmask.shape[0]));
+  int bitmask_stride0 = bitmask.strides[0];
 
   XGRAMMAR_CHECK(
       vocab_size <= bitmask_shape.second * DynamicBitset::BITS_PER_BLOCK &&
@@ -133,18 +140,18 @@ void ApplyTokenBitmaskInplaceCPU(
   // Apply mask
   if (indices.has_value()) {
     for (auto idx : indices.value()) {
-      uint32_t* data_ptr = reinterpret_cast<uint32_t*>(bitmask.data) + idx * bitmask_shape.second;
+      uint32_t* data_ptr = reinterpret_cast<uint32_t*>(bitmask.data) + idx * bitmask_stride0;
       DynamicBitset bitset(vocab_size, data_ptr);
-      auto logits_ptr = reinterpret_cast<float*>(logits->data) + idx * logits_shape.second;
+      auto logits_ptr = reinterpret_cast<float*>(logits->data) + idx * logits_stride0;
       for (int i = bitset.FindFirstZero(); i != -1; i = bitset.FindNextZero(i)) {
         logits_ptr[i] = -std::numeric_limits<float>::infinity();
       }
     }
   } else {
     for (int idx = 0; idx < logits_shape.first; ++idx) {
-      uint32_t* data_ptr = reinterpret_cast<uint32_t*>(bitmask.data) + idx * bitmask_shape.second;
+      uint32_t* data_ptr = reinterpret_cast<uint32_t*>(bitmask.data) + idx * bitmask_stride0;
       DynamicBitset bitset(vocab_size, data_ptr);
-      auto logits_ptr = reinterpret_cast<float*>(logits->data) + idx * logits_shape.second;
+      auto logits_ptr = reinterpret_cast<float*>(logits->data) + idx * logits_stride0;
       for (int i = bitset.FindFirstZero(); i != -1; i = bitset.FindNextZero(i)) {
         logits_ptr[i] = -std::numeric_limits<float>::infinity();
       }
@@ -257,14 +264,14 @@ class GrammarMatcher::Impl : public EarleyParser {
       const CompiledGrammar& compiled_grammar,
       std::optional<std::vector<int>> override_stop_tokens = std::nullopt,
       bool terminate_without_stop_token = false,
-      int max_rollback_tokens = 0
+      // max_rollback_tokens_ is deprecated and not used.
+      int max_rollback_tokens = -1
   )
       : EarleyParser(compiled_grammar->grammar, ParserState::GetInvalidState()),
         compiled_grammar_(compiled_grammar),
         tokenizer_info_(compiled_grammar->tokenizer_info),
         stop_token_ids_(override_stop_tokens.value_or(tokenizer_info_.GetStopTokenIds())),
         terminate_without_stop_token_(terminate_without_stop_token),
-        max_rollback_tokens_(max_rollback_tokens),
         tmp_accepted_bitset_(tokenizer_info_.GetVocabSize()) {
     XGRAMMAR_CHECK(!override_stop_tokens.has_value() || !override_stop_tokens->empty())
         << "The override_stop_tokens should not be empty";
@@ -284,7 +291,7 @@ class GrammarMatcher::Impl : public EarleyParser {
 
   void Reset() { EarleyParser::Reset(); }
 
-  int GetMaxRollbackTokens() const { return max_rollback_tokens_; }
+  int GetMaxRollbackTokens() const { return -1; }
 
   const std::vector<int>& GetStopTokenIds() const { return stop_token_ids_; }
 
@@ -292,6 +299,7 @@ class GrammarMatcher::Impl : public EarleyParser {
 
  private:
   using StoreType = AdaptiveTokenMask::StoreType;
+
   /*!
    * \brief If is_uncertain_saved is true, find the next token in uncertain_indices. Otherwise,
    * find the next token that is set to true in uncertain_tokens_bitset.
@@ -332,7 +340,6 @@ class GrammarMatcher::Impl : public EarleyParser {
   TokenizerInfo tokenizer_info_;
   std::vector<int> stop_token_ids_;
   bool terminate_without_stop_token_;
-  int max_rollback_tokens_;
   std::deque<int> token_length_history;
 
   // Temporary data for FillNextTokenBitmask. They are stored here to avoid repeated allocation.
@@ -341,6 +348,52 @@ class GrammarMatcher::Impl : public EarleyParser {
   std::vector<int32_t> tmp_rejected_indices_delta_;
 };
 
+class BatchGrammarMatcher::Impl {
+ public:
+  Impl(std::variant<std::string, int32_t> max_threads) {
+    if (std::holds_alternative<int32_t>(max_threads)) {
+      int32_t num_threads = std::get<int32_t>(max_threads);
+      XGRAMMAR_CHECK(num_threads >= 1)
+          << "The num_threads should be at least 1, but got " << num_threads;
+      if (num_threads > 1) {
+        if (num_threads > static_cast<int32_t>(std::thread::hardware_concurrency())) {
+          XGRAMMAR_LOG(WARNING) << "The num_threads " << num_threads << " is larger than the "
+                                << "number of hardware threads. Using "
+                                << static_cast<int32_t>(std::thread::hardware_concurrency())
+                                << " instead.";
+        }
+        max_threads_ =
+            std::min(num_threads, static_cast<int32_t>(std::thread::hardware_concurrency()));
+      }
+    } else {
+      std::string str = std::get<std::string>(max_threads);
+      XGRAMMAR_CHECK(str == "auto");
+      max_threads_ = std::thread::hardware_concurrency() / 2;
+    }
+  }
+
+  void BatchFillNextTokenBitmask(
+      std::vector<GrammarMatcher>* matchers,
+      DLTensor* next_token_bitmask,
+      const std::optional<std::vector<int32_t>>& indices,
+      bool debug_print
+  );
+
+  static std::vector<uint8_t> BatchAcceptToken(
+      std::vector<GrammarMatcher>* matchers, const std::vector<int32_t>& token_ids, bool debug_print
+  );
+
+  static std::vector<uint8_t> BatchAcceptString(
+      std::vector<GrammarMatcher>* matchers,
+      const std::vector<std::string>& input_strs,
+      bool debug_print
+  );
+
+ private:
+  std::optional<ThreadPool> thread_pool_ = std::nullopt;
+  int32_t max_threads_ = 1;
+};
+
 bool GrammarMatcher::Impl::AcceptStopToken() {
   if (terminate_without_stop_token_) {
     return false;
@@ -383,8 +436,8 @@ bool GrammarMatcher::Impl::AcceptToken(int32_t token_id, bool debug_print) {
       states_str += "  " + state.ToString() + "\n";
     }
     XGRAMMAR_LOG(INFO) << "Accepting token id " << token_id << ", string: \""
-                       << PrintAsEscapedUTF8(tokenizer_info_.GetDecodedVocab()[token_id])
-                       << "\", current scannable states:\n"
+                       << EscapeString(tokenizer_info_.GetDecodedVocab()[token_id])
+                       << "\", current state:\n"
                        << states_str;
   }
   // Handle the stop token
@@ -409,11 +462,11 @@ bool GrammarMatcher::Impl::AcceptToken(int32_t token_id, bool debug_print) {
   const auto& token = tokenizer_info_.GetDecodedVocab()[token_id];
   int pos = 0;
   for (auto char_value : token) {
-    if (!Advance(char_value)) {
+    if (!Advance(char_value, debug_print)) {
       if (debug_print) {
-        XGRAMMAR_LOG(INFO) << "Token #" << token_id << "<" << PrintAsEscapedUTF8(token)
+        XGRAMMAR_LOG(INFO) << "Token #" << token_id << "<" << EscapeString(token)
                            << "> rejected at position " << pos << ", char "
-                           << PrintAsEscapedUTF8(char_value);
+                           << EscapeString(char_value);
       }
       PopLastStates(pos);
       return false;
@@ -421,13 +474,10 @@ bool GrammarMatcher::Impl::AcceptToken(int32_t token_id, bool debug_print) {
     ++pos;
   }
   token_length_history.push_back(token.size());
-  if (static_cast<int>(token_length_history.size()) > max_rollback_tokens_) {
-    token_length_history.pop_front();
-  }
 
   if (debug_print) {
     XGRAMMAR_LOG(INFO) << "Token #" << token_id << "<"
-                       << PrintAsEscapedUTF8(tokenizer_info_.GetDecodedVocab()[token_id])
+                       << EscapeString(tokenizer_info_.GetDecodedVocab()[token_id])
                        << "> accepted.";
   }
   return true;
@@ -436,36 +486,36 @@ bool GrammarMatcher::Impl::AcceptToken(int32_t token_id, bool debug_print) {
 bool GrammarMatcher::Impl::AcceptString(const std::string& input_str, bool debug_print) {
   if (IsStopTokenAccepted()) {
     XGRAMMAR_LOG(WARNING) << "The matcher has terminated after accepting the stop token, but is "
-                          << "trying to accept new string \"" << PrintAsEscapedUTF8(input_str)
-                          << "\".";
+                          << "trying to accept new string \"" << EscapeString(input_str) << "\".";
     return false;
   }
 
+  if (debug_print) {
+    XGRAMMAR_LOG(INFO) << "Trying to accept string \"" << EscapeString(input_str)
+                       << "\". Current state:\n"
+                       << PrintStates();
+  }
+
   int accepted_cnt = 0;
   for (auto char_value : input_str) {
-    if (!Advance(char_value)) {
+    if (!Advance(char_value, debug_print)) {
       if (debug_print) {
-        XGRAMMAR_LOG(INFO) << "String \"" << PrintAsEscapedUTF8(input_str)
-                           << "\" rejected at position " << accepted_cnt << ", char "
-                           << PrintAsEscapedUTF8(char_value);
+        XGRAMMAR_LOG(INFO) << "String \"" << EscapeString(input_str) << "\" is rejected at "
+                           << "position " << accepted_cnt << ", char " << EscapeString(char_value);
       }
       PopLastStates(accepted_cnt);
       return false;
     }
+    if (debug_print) {
+      XGRAMMAR_LOG(INFO) << "Char " << EscapeString(char_value) << " is accepted. Current state:\n"
+                         << PrintStates();
+    }
     ++accepted_cnt;
   }
   token_length_history.push_back(input_str.size());
-  if (static_cast<int>(token_length_history.size()) > max_rollback_tokens_) {
-    token_length_history.pop_front();
-  }
+
   if (debug_print) {
-    std::string states_str;
-    for (const auto& state : GetLatestScanableStates()) {
-      states_str += "  " + state.ToString() + "\n";
-    }
-    XGRAMMAR_LOG(INFO) << "String \"" << PrintAsEscapedUTF8(input_str)
-                       << "\" is accepted. Current scannable states:\n"
-                       << states_str;
+    XGRAMMAR_LOG(INFO) << "String \"" << EscapeString(input_str) << "\" is accepted.";
   }
   return true;
 }
@@ -525,10 +575,6 @@ bool GrammarMatcher::Impl::FillNextTokenBitmask(
   // {-1} means the universal set, i.e. all tokens initially
   tmp_rejected_indices_.assign({-1});
 
-  // If there is a leaf ParserState that is a tag dispatch, we allow special tokens to be accepted
-  // because in function calling cases, only the part within the tag is constrained
-  bool have_tag_dispatch = false;
-
   if (debug_print) {
     XGRAMMAR_LOG(INFO) << "FillNextTokenBitmask: index=" << index
                        << ", num of states=" << latest_states.size();
@@ -538,14 +584,6 @@ bool GrammarMatcher::Impl::FillNextTokenBitmask(
       latest_states_with_masks;
 
   for (const auto& state : latest_states) {
-    auto cur_sequence = grammar_->GetGrammarExpr(state.sequence_id);
-    XGRAMMAR_DCHECK(
-        !(cur_sequence.type == GrammarExprType::kRuleRef ||
-          cur_sequence.type == GrammarExprType::kChoices ||
-          cur_sequence.type == GrammarExprType::kEmptyStr)
-    );
-    have_tag_dispatch = cur_sequence.type == GrammarExprType::kTagDispatch;
-    XGRAMMAR_DCHECK(have_tag_dispatch || cur_sequence.type == GrammarExprType::kSequence);
     auto adaptive_token_mask_it = adaptive_token_mask_cache.find(state);
     XGRAMMAR_CHECK(adaptive_token_mask_it != adaptive_token_mask_cache.end()) << state;
     const auto& adaptive_token_mask = adaptive_token_mask_it->second;
@@ -659,11 +697,7 @@ bool GrammarMatcher::Impl::FillNextTokenBitmask(
   // Finally update the rejected_ids bitset
   bool can_reach_end = IsCompleted();
   SetTokenBitmask(
-      bitmask_data_ptr,
-      tmp_accepted_bitset_,
-      tmp_rejected_indices_,
-      can_reach_end,
-      have_tag_dispatch
+      bitmask_data_ptr, tmp_accepted_bitset_, tmp_rejected_indices_, can_reach_end, false
   );
   if (debug_print) {
     XGRAMMAR_LOG(INFO) << "Filled bitmask: " << PrintBitmask(bitmask_data_ptr, tokenizer_info_);
@@ -683,32 +717,56 @@ std::string GrammarMatcher::Impl::FindJumpForwardString() {
   while (can_find_next_char) {
     const auto& states = scanable_state_history_[scanable_state_history_.size() - 1];
 
+    // The state comes to the end of the grammar
+    if (IsCompleted()) {
+      can_find_next_char = false;
+      break;
+    }
+
     // 1. Check that for every leaf ParserState, the next possible char is unique and the same
     // -1 means not found yet; 0~255 means the next char
     int next_char = -1;
-    for (const auto& ParserState : states) {
-      if (IsCompleted()) {
-        can_find_next_char = false;
-        break;
+    for (const auto& state : states) {
+      if (state.rule_id != -1 && grammar_->per_rule_fsms[state.rule_id].has_value()) {
+        const auto& fsm = grammar_->per_rule_fsms[state.rule_id].value();
+        const auto& current_edges = fsm.GetFsm().GetEdges(state.element_id);
+        for (const auto& edge : current_edges) {
+          if (!edge.IsCharRange()) {
+            continue;
+          }
+          if (edge.min != edge.max) {
+            can_find_next_char = false;
+            break;
+          }
+          if (next_char == -1) {
+            next_char = edge.min;
+          } else if (next_char != edge.min) {
+            can_find_next_char = false;
+            break;
+          }
+        }
+        continue;
       }
-      auto cur_sequence = grammar_->GetGrammarExpr(ParserState.sequence_id);
+
+      auto cur_sequence = grammar_->GetGrammarExpr(state.sequence_id);
+
       // We cannot deduce the next char for tag dispatch
       if (cur_sequence.type == GrammarExprType::kTagDispatch) {
         can_find_next_char = false;
         break;
       }
       // The ParserState comes to the end of the grammar
-      XGRAMMAR_DCHECK(ParserState.element_id != cur_sequence.size());
+      XGRAMMAR_DCHECK(state.element_id != cur_sequence.size());
       XGRAMMAR_DCHECK(
           cur_sequence.type != GrammarExprType::kChoices &&
           cur_sequence.type != GrammarExprType::kEmptyStr
       );
-      const auto& cur_element = grammar_->GetGrammarExpr(cur_sequence[ParserState.element_id]);
+      const auto& cur_element = grammar_->GetGrammarExpr(cur_sequence[state.element_id]);
       if (cur_element.type == GrammarExprType::kByteString) {
-        XGRAMMAR_DCHECK(ParserState.sub_element_id < cur_element.size());
+        XGRAMMAR_DCHECK(state.sub_element_id < cur_element.size());
         if (next_char == -1) {
-          next_char = cur_element[ParserState.sub_element_id];
-        } else if (next_char != cur_element[ParserState.sub_element_id]) {
+          next_char = cur_element[state.sub_element_id];
+        } else if (next_char != cur_element[state.sub_element_id]) {
           can_find_next_char = false;
           break;
         }
@@ -722,7 +780,7 @@ std::string GrammarMatcher::Impl::FindJumpForwardString() {
           cur_element.type == GrammarExprType::kCharacterClass ||
           cur_element.type == GrammarExprType::kCharacterClassStar
       );
-      if (ParserState.sub_element_id > 0 || cur_element.size() != 3 || cur_element[0] != 0 ||
+      if (state.sub_element_id > 0 || cur_element.size() != 3 || cur_element[0] != 0 ||
           cur_element[1] != cur_element[2]) {
         can_find_next_char = false;
         break;
@@ -845,6 +903,75 @@ int GrammarMatcher::Impl::GetNextUncertainToken(
   }
 }
 
+void BatchGrammarMatcher::Impl::BatchFillNextTokenBitmask(
+    std::vector<GrammarMatcher>* matchers,
+    DLTensor* next_token_bitmask,
+    const std::optional<std::vector<int32_t>>& indices,
+    bool debug_print
+) {
+  XGRAMMAR_CHECK(!indices.has_value() || indices->size() == matchers->size())
+      << "The size of indices (" << (indices.has_value() ? indices->size() : 0)
+      << ") should be the same as the size of matchers (" << matchers->size() << ").";
+  // Initialize the thread pool if needed. It should be initialized each time,
+  // because ThreadPool cannot be reused after Join().
+  if (max_threads_ > 1) {
+    thread_pool_.emplace(max_threads_);
+  }
+  if (!thread_pool_.has_value()) {
+    for (int i = 0; i < static_cast<int32_t>(matchers->size()); i++) {
+      auto& matcher = (*matchers)[i];
+      int index = indices.has_value() ? (*indices)[i] : i;
+      XGRAMMAR_CHECK(index >= 0 && index < next_token_bitmask->shape[0])
+          << "The index " << index << " is out of range [0, " << next_token_bitmask->shape[0]
+          << ") for batch_id " << i << ".";
+      matcher->FillNextTokenBitmask(next_token_bitmask, index, debug_print);
+    }
+  } else {
+    auto fill_next_token_mask = [&](int32_t batch_id) {
+      auto& matcher = (*matchers)[batch_id];
+      int index = indices.has_value() ? (*indices)[batch_id] : batch_id;
+      XGRAMMAR_CHECK(index >= 0 && index < next_token_bitmask->shape[0])
+          << "The index " << index << " is out of range [0, " << next_token_bitmask->shape[0]
+          << ") for batch_id " << batch_id << ".";
+      matcher->FillNextTokenBitmask(next_token_bitmask, index, debug_print);
+    };
+    for (int i = 0; i < static_cast<int32_t>(matchers->size()); i++) {
+      thread_pool_->Execute([fill_next_token_mask, i]() { fill_next_token_mask(i); });
+    }
+    thread_pool_->Join();
+  }
+}
+
+std::vector<uint8_t> BatchGrammarMatcher::Impl::BatchAcceptString(
+    std::vector<GrammarMatcher>* matchers,
+    const std::vector<std::string>& input_strs,
+    bool debug_print
+) {
+  XGRAMMAR_CHECK(matchers->size() == input_strs.size())
+      << "The size of matchers (" << matchers->size() << ") and input_strs (" << input_strs.size()
+      << ") should be the same.";
+  std::vector<uint8_t> accepted(matchers->size());
+  for (int i = 0; i < static_cast<int32_t>(matchers->size()); i++) {
+    auto& matcher = (*matchers)[i];
+    accepted[i] = matcher->AcceptString(input_strs[i], debug_print);
+  }
+  return accepted;
+}
+
+std::vector<uint8_t> BatchGrammarMatcher::Impl::BatchAcceptToken(
+    std::vector<GrammarMatcher>* matchers, const std::vector<int32_t>& token_ids, bool debug_print
+) {
+  XGRAMMAR_CHECK(matchers->size() == token_ids.size())
+      << "The size of matchers (" << matchers->size() << ") and token_ids (" << token_ids.size()
+      << ") should be the same.";
+  std::vector<uint8_t> accepted(matchers->size());
+  for (int i = 0; i < static_cast<int32_t>(matchers->size()); i++) {
+    auto& matcher = (*matchers)[i];
+    accepted[i] = matcher->AcceptToken(token_ids[i], debug_print);
+  }
+  return accepted;
+}
+
 GrammarMatcher::GrammarMatcher(
     const CompiledGrammar& compiled_grammar,
     std::optional<std::vector<int>> override_stop_tokens,
@@ -887,4 +1014,30 @@ std::string GrammarMatcher::_DebugPrintInternalState() const {
   return pimpl_->_DebugPrintInternalState();
 }
 
+void BatchGrammarMatcher::BatchFillNextTokenBitmask(
+    std::vector<GrammarMatcher>* matchers,
+    DLTensor* next_token_bitmask,
+    const std::optional<std::vector<int32_t>>& indices,
+    bool debug_print
+) {
+  return pimpl_->BatchFillNextTokenBitmask(matchers, next_token_bitmask, indices, debug_print);
+}
+
+std::vector<uint8_t> BatchGrammarMatcher::BatchAcceptString(
+    std::vector<GrammarMatcher>* matchers,
+    const std::vector<std::string>& input_strs,
+    bool debug_print
+) {
+  return Impl::BatchAcceptString(matchers, input_strs, debug_print);
+}
+
+std::vector<uint8_t> BatchGrammarMatcher::BatchAcceptToken(
+    std::vector<GrammarMatcher>* matchers, const std::vector<int32_t>& token_ids, bool debug_print
+) {
+  return Impl::BatchAcceptToken(matchers, token_ids, debug_print);
+}
+
+BatchGrammarMatcher::BatchGrammarMatcher(std::variant<std::string, int32_t> max_threads)
+    : pimpl_(std::make_shared<BatchGrammarMatcher::Impl>(max_threads)) {}
+
 }  // namespace xgrammar
