@@ -7,12 +7,17 @@
 
 #include <picojson.h>
 
+#include <cstdint>
+#include <optional>
+#include <string>
 #include <variant>
+#include <vector>
 
 #include "grammar_builder.h"
-#include "grammar_data_structure.h"
+#include "grammar_impl.h"
 #include "support/encoding.h"
 #include "support/logging.h"
+#include "xgrammar/grammar.h"
 
 namespace xgrammar {
 
@@ -173,7 +178,7 @@ EBNFLexer::Token EBNFLexer::Impl::ParseStringToken() {
   // Convert codepoints to UTF-8 string value
   std::string value;
   for (auto codepoint : codepoints) {
-    value += PrintAsUTF8(codepoint);
+    value += CharToUTF8(codepoint);
   }
 
   return {TokenType::StringLiteral, lexeme, value, start_line, start_column};
@@ -622,12 +627,7 @@ int32_t EBNFParser::ParseString() {
     return builder_.AddEmptyStr();
   }
 
-  // Convert string to bytes
-  std::vector<int32_t> bytes;
-  for (auto c : str_value) {
-    bytes.push_back(static_cast<int32_t>(static_cast<uint8_t>(c)));
-  }
-  return builder_.AddByteString(bytes);
+  return builder_.AddByteString(str_value);
 }
 
 int32_t EBNFParser::ParseRuleRef() {
@@ -757,58 +757,130 @@ int32_t EBNFParser::HandleQuestionQuantifier(int32_t grammar_expr_id) {
   return builder_.AddRuleRef(new_rule_id);
 }
 
-int32_t EBNFParser::HandleRepetitionRange(int32_t grammar_expr_id, int64_t lower, int64_t upper) {
-  // Construct expr expr ... expr (l times)
-  std::vector<int32_t> elements;
-  for (int64_t i = 0; i < lower; ++i) {
-    elements.push_back(grammar_expr_id);
+int32_t EBNFParser::HandleRepetitionRange(
+    const int32_t grammar_expr_id, int64_t lower, int64_t upper
+) {
+  static const int64_t kUnzipThreshold = 15;
+  XGRAMMAR_DCHECK(lower >= 0);
+  XGRAMMAR_DCHECK(upper == -1 || upper >= lower);
+  // Case 1. small (<=threshold), unzip the repetition.
+  if (upper != -1 && upper <= kUnzipThreshold) {
+    std::vector<int32_t> choices;
+    if (lower == 0) {
+      choices.push_back(builder_.AddEmptyStr());
+      lower = 1;  // We have already handled the empty string case.
+    }
+    for (int64_t count = lower; count <= upper; ++count) {
+      std::vector<int32_t> sequence;
+      for (int64_t i = 0; i < count; ++i) {
+        sequence.push_back(grammar_expr_id);
+      }
+      choices.push_back(builder_.AddSequence(sequence));
+    }
+    return builder_.AddChoices(choices);
   }
 
-  // Case 1: {l}:
-  // expr expr ... expr (l times)
-  if (upper == lower) {
-    return builder_.AddSequence(elements);
+  // Case 2. upper is unbounded or large.
+
+  // Case 2.1.1. lower is smaller than threshold, and upper is large. Transform {lower, upper} into:
+  // {threshold, upper} | {lower} | ... | {threshold}.
+  // Case 2.1.2. lower is smaller than threshold, and upper is unbounded. Unzip the
+  // {lower} repetition, and add a star expression.
+  std::vector<int32_t> choices;
+  if (lower < kUnzipThreshold) {
+    if (upper == -1) {
+      int infinite_repetition_id = -1;
+      const auto& rule_expr = builder_.GetGrammarExpr(grammar_expr_id);
+      if (rule_expr.type == GrammarBuilder::GrammarExprType::kCharacterClass) {
+        std::vector<GrammarBuilder::CharacterClassElement> character_ranges;
+        bool is_negative = rule_expr[0];
+        for (int i = 1; i < static_cast<int>(rule_expr.size()); i += 2) {
+          character_ranges.push_back({rule_expr[i], rule_expr[i + 1]});
+        }
+        infinite_repetition_id = builder_.AddCharacterClassStar(character_ranges, is_negative);
+      } else {
+        const auto& unbounded_rule_id =
+            builder_.AddEmptyRule(builder_.GetNewRuleName(cur_rule_name_ + "_repeat_inf"));
+        int recursion_sequence =
+            builder_.AddSequence({grammar_expr_id, builder_.AddRuleRef(unbounded_rule_id)});
+        int recursion_choice = builder_.AddChoices({builder_.AddEmptyStr(), recursion_sequence});
+        builder_.UpdateRuleBody(unbounded_rule_id, recursion_choice);
+        infinite_repetition_id = builder_.AddRuleRef(unbounded_rule_id);
+      }
+      std::vector<int32_t> sequence(lower, grammar_expr_id);
+      sequence.push_back(infinite_repetition_id);
+      return builder_.AddSequence(sequence);
+    }
+    if (lower == 0) {
+      choices.push_back(builder_.AddEmptyStr());
+      lower = 1;
+    }
+    for (; lower < kUnzipThreshold; ++lower) {
+      std::vector<int32_t> sequence;
+      for (int64_t i = 0; i < lower; ++i) {
+        sequence.push_back(grammar_expr_id);
+      }
+      choices.push_back(builder_.AddSequence(sequence));
+    }
   }
 
-  // Case 2: {l,}:
-  // expr expr ... expr (l times) rest
-  // rest ::= "" | expr rest
+  std::optional<int32_t> infinite_repetition_id = std::nullopt;
+  std::vector<int32_t> repeated_sequence;
+  // Now, we transform {lower, upper} into {max{threshold, lower}, upper}.
+  // Case 2.2 upper is unbounded. We will transform it into {lower} {0, inf}.
   if (upper == -1) {
-    auto new_rule_name = builder_.GetNewRuleName(cur_rule_name_);
-    auto new_rule_id = builder_.AddEmptyRule(new_rule_name);
-    auto ref_to_new_rule = builder_.AddRuleRef(new_rule_id);
-    auto new_grammar_expr_id = builder_.AddChoices(
-        {builder_.AddEmptyStr(), builder_.AddSequence({grammar_expr_id, ref_to_new_rule})}
-    );
-    builder_.UpdateRuleBody(new_rule_id, new_grammar_expr_id);
-    elements.push_back(builder_.AddRuleRef(new_rule_id));
-    return builder_.AddSequence(elements);
+    const auto& rule_expr = builder_.GetGrammarExpr(grammar_expr_id);
+    if (rule_expr.type == GrammarBuilder::GrammarExprType::kCharacterClass) {
+      std::vector<GrammarBuilder::CharacterClassElement> character_ranges;
+      bool is_negative = rule_expr[0];
+      for (int i = 1; i < static_cast<int>(rule_expr.size()); i += 2) {
+        character_ranges.push_back({rule_expr[i], rule_expr[i + 1]});
+      }
+      infinite_repetition_id = builder_.AddCharacterClassStar(character_ranges, is_negative);
+    } else {
+      const auto& unbounded_rule_id =
+          builder_.AddEmptyRule(builder_.GetNewRuleName(cur_rule_name_ + "_repeat_inf"));
+      int recursion_sequence =
+          builder_.AddSequence({grammar_expr_id, builder_.AddRuleRef(unbounded_rule_id)});
+      int recursion_choice = builder_.AddChoices({builder_.AddEmptyStr(), recursion_sequence});
+      builder_.UpdateRuleBody(unbounded_rule_id, recursion_choice);
+      infinite_repetition_id = builder_.AddRuleRef(unbounded_rule_id);
+    }
+    upper = lower;
   }
 
-  // Case 3: {l, r} (r - l >= 1)
-  // expr expr ... expr (l times) rest1
-  // rest1 ::= "" | expr rest2
-  // rest2 ::= "" | expr rest3
-  // ...
-  // rest(r - l) ::= "" | expr
-  std::vector<int32_t> rest_rule_ids;
+  // Handle the {lower, upper} part, where threshold <= lower <= upper.
+  const auto repeat_name = cur_rule_name_ + "_repeat_1";
+  XGRAMMAR_DCHECK(lower >= kUnzipThreshold && upper >= lower);
 
-  for (int64_t i = 0; i < upper - lower; ++i) {
-    auto new_rule_name = builder_.GetNewRuleName(cur_rule_name_);
-    rest_rule_ids.push_back(builder_.AddEmptyRule(new_rule_name));
+  // If we have infinite repetition part, add it to the sequence.
+  if (infinite_repetition_id.has_value()) {
+    repeated_sequence.push_back(infinite_repetition_id.value());
   }
-  for (int64_t i = 0; i < upper - lower - 1; ++i) {
-    auto ref_to_next_rule = builder_.AddRuleRef(rest_rule_ids[i + 1]);
-    auto new_grammar_expr_id = builder_.AddChoices(
-        {builder_.AddEmptyStr(), builder_.AddSequence({grammar_expr_id, ref_to_next_rule})}
-    );
-    builder_.UpdateRuleBody(rest_rule_ids[i], new_grammar_expr_id);
-  }
-  auto last_grammar_expr_id = builder_.AddChoices({builder_.AddEmptyStr(), grammar_expr_id});
-  builder_.UpdateRuleBody(rest_rule_ids.back(), last_grammar_expr_id);
 
-  elements.push_back(builder_.AddRuleRef(rest_rule_ids[0]));
-  return builder_.AddSequence(elements);
+  // The repetition body.
+  if (upper != kUnzipThreshold) {
+    XGRAMMAR_DCHECK(upper > kUnzipThreshold);
+    auto new_grammar_expr_id = builder_.AddChoices({builder_.AddSequence({grammar_expr_id})});
+    auto new_rule_id = builder_.AddRuleWithHint(repeat_name, new_grammar_expr_id);
+    auto new_repeated_ref_rule_expr = builder_.AddChoices({builder_.AddSequence(
+        {builder_.AddRepeat(new_rule_id, lower - kUnzipThreshold, upper - kUnzipThreshold)}
+    )});
+    auto new_repeated_rule_id =
+        builder_.AddRuleWithHint(repeat_name + "_inner", new_repeated_ref_rule_expr);
+    repeated_sequence.push_back(builder_.AddRuleRef(new_repeated_rule_id));
+    std::vector<int32_t> repetition_lookahead(kUnzipThreshold, grammar_expr_id);
+    builder_.UpdateLookaheadAssertion(new_rule_id, builder_.AddSequence(repetition_lookahead));
+  }
+
+  // Add the last threshold grammar_expr_id to the sequence.
+  for (int i = 0; i < kUnzipThreshold; ++i) {
+    repeated_sequence.push_back(grammar_expr_id);
+  }
+
+  // Add the sequence to choices.
+  choices.push_back(builder_.AddSequence(repeated_sequence));
+  return builder_.AddChoices(choices);
 }
 
 int32_t EBNFParser::ParseElementWithQuantifier() {
@@ -948,10 +1020,10 @@ int32_t EBNFParser::ParseTagDispatch() {
   auto start = current_token_;
   auto args = ParseMacroArguments();
   auto delta_element = start - current_token_;  // Used to report parse errors
-  // Process the arguments for TagDispatch
-  std::vector<std::pair<int32_t, int32_t>> tag_rule_pairs;
 
-  // Process each argument in the form of ("tag", rule_name)
+  Grammar::Impl::TagDispatch tag_dispatch;
+
+  // Position parameters: ("tag", rule_name)
   for (const auto& arg : args.arguments) {
     auto tuple_node = std::get_if<MacroIR::TupleNode>(arg.get());
     if (tuple_node == nullptr) {
@@ -967,8 +1039,6 @@ int32_t EBNFParser::ParseTagDispatch() {
     if (tag_str_node == nullptr || tag_str_node->value.empty()) {
       ReportParseError("Tag must be a non-empty string literal", delta_element);
     }
-    auto tag_id = builder_.AddByteString(tag_str_node->value);
-
     // Second element should be an identifier (rule name)
     auto rule_name_node = std::get_if<MacroIR::IdentifierNode>(tuple_node->elements[1].get());
     if (rule_name_node == nullptr) {
@@ -980,10 +1050,54 @@ int32_t EBNFParser::ParseTagDispatch() {
       ReportParseError("Rule \"" + rule_name_node->name + "\" is not defined", delta_element);
     }
 
-    tag_rule_pairs.push_back({tag_id, rule_id});
+    tag_dispatch.tag_rule_pairs.push_back({tag_str_node->value, rule_id});
   }
 
-  return builder_.AddTagDispatch(tag_rule_pairs);
+  // stop_eos
+  tag_dispatch.stop_eos = true;
+  if (auto it = args.named_arguments.find("stop_eos"); it != args.named_arguments.end()) {
+    auto bool_node = std::get_if<MacroIR::BooleanNode>(it->second.get());
+    if (bool_node == nullptr) {
+      ReportParseError("stop_eos must be a boolean literal", delta_element);
+    }
+    tag_dispatch.stop_eos = bool_node->value;
+  }
+
+  // stop_str
+  if (auto it = args.named_arguments.find("stop_str"); it != args.named_arguments.end()) {
+    auto tuple_node = std::get_if<MacroIR::TupleNode>(it->second.get());
+    if (tuple_node == nullptr) {
+      ReportParseError("Stop strings must be a tuple", delta_element);
+    }
+
+    for (const auto& element : tuple_node->elements) {
+      auto stop_str_node = std::get_if<MacroIR::StringNode>(element.get());
+      if (stop_str_node == nullptr || stop_str_node->value.empty()) {
+        ReportParseError("Stop string must be a non-empty string literal", delta_element);
+      }
+      tag_dispatch.stop_str.push_back(stop_str_node->value);
+    }
+  }
+
+  // loop_after_dispatch
+  tag_dispatch.loop_after_dispatch = true;
+  if (auto it = args.named_arguments.find("loop_after_dispatch");
+      it != args.named_arguments.end()) {
+    auto bool_node = std::get_if<MacroIR::BooleanNode>(it->second.get());
+    if (bool_node == nullptr) {
+      ReportParseError("loop_after_dispatch must be a boolean literal", delta_element);
+    }
+    tag_dispatch.loop_after_dispatch = bool_node->value;
+  }
+
+  // Well formed check
+  if (!tag_dispatch.stop_eos && tag_dispatch.stop_str.empty()) {
+    ReportParseError(
+        "The TagDispatch must have stop_eos=true or stop_str is not empty", delta_element
+    );
+  }
+
+  return builder_.AddTagDispatch(tag_dispatch);
 }
 
 int32_t EBNFParser::ParseLookaheadAssertion() {
