@@ -23,10 +23,12 @@ namespace xgrammar {
 class RegexConverter {
  public:
   explicit RegexConverter(const std::string& regex) : regex_(regex) {
-    regex_codepoints_ = ParseUTF8(regex_.c_str(), false);
-    if (regex_codepoints_[0] == kInvalidUTF8) {
-      XGRAMMAR_LOG(FATAL) << "The regex is not a valid UTF-8 string.";
-      XGRAMMAR_UNREACHABLE();
+    if (!regex.empty()) {
+      regex_codepoints_ = ParseUTF8(regex_.c_str(), false);
+      if (regex_codepoints_[0] == kInvalidUTF8) {
+        XGRAMMAR_LOG(FATAL) << "The regex is not a valid UTF-8 string.";
+        XGRAMMAR_UNREACHABLE();
+      }
     }
     regex_codepoints_.push_back(0);  // Add a null terminator
   }
@@ -90,7 +92,7 @@ std::string RegexConverter::HandleCharacterClass() {
     if (*current_ == '\\') {
       char_class += HandleEscapeInCharClass();
     } else {
-      char_class += PrintAsUTF8(*current_);
+      char_class += CharToUTF8(*current_);
       ++current_;
     }
   }
@@ -154,7 +156,7 @@ std::string RegexConverter::HandleCharEscape() {
   auto [codepoint, len] = ParseNextEscaped(current_, CUSTOM_ESCAPE_MAP);
   if (codepoint != CharHandlingError::kInvalidEscape) {
     current_ += len;
-    return PrintAsEscapedUTF8(codepoint);
+    return EscapeString(codepoint);
   } else if (current_[1] == 'u' && current_[2] == '{') {
     current_ += 3;
     int len = 0;
@@ -167,21 +169,21 @@ std::string RegexConverter::HandleCharEscape() {
       RaiseError("Invalid Unicode escape sequence.");
     }
     current_ += len + 1;
-    return PrintAsEscapedUTF8(value);
+    return EscapeString(value);
   } else if (current_[1] == 'c') {
     current_ += 2;
     if (!std::isalpha(*current_)) {
       RaiseError("Invalid control character escape sequence.");
     }
     ++current_;
-    return PrintAsEscapedUTF8((*(current_ - 1)) % 32);
+    return EscapeString((*(current_ - 1)) % 32);
   } else {
     RaiseWarning(
-        "Escape sequence '\\" + PrintAsEscapedUTF8(current_[1]) +
+        "Escape sequence '\\" + EscapeString(current_[1]) +
         "' is not recognized. The character itself will be matched"
     );
     current_ += 2;
-    return PrintAsEscapedUTF8(current_[-1]);
+    return EscapeString(current_[-1]);
   }
 }
 
@@ -291,6 +293,7 @@ std::string RegexConverter::Convert() {
   start_ = regex_codepoints_.data();
   current_ = start_;
   end_ = start_ + regex_codepoints_.size() - 1;
+  bool is_empty = true;
   while (current_ != end_) {
     if (*current_ == '^') {
       if (current_ != start_) {
@@ -307,8 +310,10 @@ std::string RegexConverter::Convert() {
       }
       ++current_;
     } else if (*current_ == '[') {
+      is_empty = false;
       AddEBNFSegment(HandleCharacterClass());
     } else if (*current_ == '(') {
+      is_empty = false;
       ++current_;
       ++parenthesis_level_;
       AddEBNFSegment("(");
@@ -317,6 +322,7 @@ std::string RegexConverter::Convert() {
         HandleGroupModifier();
       }
     } else if (*current_ == ')') {
+      is_empty = false;
       if (parenthesis_level_ == 0) {
         RaiseError("Unmatched ')'");
       }
@@ -328,6 +334,7 @@ std::string RegexConverter::Convert() {
       AddEBNFSegment(")");
       ++current_;
     } else if (*current_ == '*' || *current_ == '+' || *current_ == '?') {
+      is_empty = false;
       result_ebnf_ += static_cast<char>(*current_);
       ++current_;
       if (current_ != end_ && *current_ == '?') {
@@ -340,6 +347,7 @@ std::string RegexConverter::Convert() {
         RaiseError("Two consecutive repetition modifiers are not allowed.");
       }
     } else if (*current_ == '{') {
+      is_empty = false;
       result_ebnf_ += HandleRepetitionRange();
       if (current_ != end_ && *current_ == '?') {
         // Still ignore the non-greedy modifier.
@@ -350,22 +358,29 @@ std::string RegexConverter::Convert() {
         RaiseError("Two consecutive repetition modifiers are not allowed.");
       }
     } else if (*current_ == '|') {
+      is_empty = false;
       AddEBNFSegment("|");
       ++current_;
     } else if (*current_ == '\\') {
+      is_empty = false;
       AddEBNFSegment(HandleEscape());
     } else if (*current_ == '.') {
+      is_empty = false;
       AddEBNFSegment(R"([\u0000-\U0010FFFF])");
       ++current_;
     } else {
+      is_empty = false;
       // Non-special characters are matched literally.
-      AddEBNFSegment("\"" + PrintAsEscapedUTF8(*current_) + "\"");
+      AddEBNFSegment("\"" + EscapeString(*current_) + "\"");
       ++current_;
     }
   }
   if (parenthesis_level_ != 0) {
     RaiseError("The parenthesis is not closed.");
   }
+  if (is_empty) {
+    AddEBNFSegment("\"\"");
+  }
   return result_ebnf_;
 }
 
