@@ -8,6 +8,7 @@
 #include <string>
 #include <vector>
 
+#include "grammar_impl.h"
 #include "grammar_parser.h"
 #include "support/encoding.h"
 
@@ -21,8 +22,7 @@ std::string PrintTokenByIds(
   ss << "[";
   int print_num = std::min(static_cast<int>(token_ids.size()), max_print_num);
   for (int i = 0; i < print_num; ++i) {
-    ss << "#" << token_ids[i] << " <" << PrintAsEscapedUTF8(sorted_decoded_vocab[token_ids[i]])
-       << ">";
+    ss << "#" << token_ids[i] << " <" << EscapeString(sorted_decoded_vocab[token_ids[i]]) << ">";
     if (i < print_num - 1) {
       ss << ", ";
     }
@@ -40,4 +40,18 @@ Grammar _EBNFToGrammarNoNormalization(
   return ParseEBNF(ebnf_string, root_rule_name);
 }
 
+std::string _PrintGrammarFSMs(const Grammar& grammar) {
+  std::string result;
+  for (int i = 0; i < grammar->NumRules(); i++) {
+    result += "Rule " + std::to_string(i) + ": " + grammar->GetRule(i).name + ", FSM: ";
+    if (grammar->per_rule_fsms[i].has_value()) {
+      result += grammar->per_rule_fsms[i]->ToString();
+    } else {
+      result += "None";
+    }
+    result += "\n";
+  }
+  return result;
+}
+
 }  // namespace xgrammar
