@@ -12,9 +12,13 @@
 
 #include <cstddef>
 #include <optional>
+#include <stdexcept>
 #include <string>
+#include <variant>
 #include <vector>
 
+#include "xgrammar/exception.h"
+
 namespace xgrammar {
 
 /*!
@@ -23,8 +27,12 @@ namespace xgrammar {
  */
 class CompiledGrammar {
  public:
+  /*! \brief Get the associated grammar. */
   Grammar GetGrammar() const;
+
+  /*! \brief Get the associated tokenizer info. */
   TokenizerInfo GetTokenizerInfo() const;
+
   /*! \brief Return the approximate memory usage of the grammar in bytes. */
   std::size_t MemorySizeBytes() const;
 
@@ -32,8 +40,8 @@ class CompiledGrammar {
   std::string SerializeJSON() const;
 
   /*! \brief Deserialize a compiled grammar from a JSON string and tokenizer info. */
-  static CompiledGrammar DeserializeJSON(
-      const std::string& json_string, TokenizerInfo tokenizer_info
+  static std::variant<CompiledGrammar, SerializationError> DeserializeJSON(
+      const std::string& json_string, const TokenizerInfo& tokenizer_info
   );
 
   XGRAMMAR_DEFINE_PIMPL_METHODS(CompiledGrammar);
@@ -60,7 +68,7 @@ class GrammarCompiler {
       const TokenizerInfo& tokenizer_info,
       int max_threads = 8,
       bool cache_enabled = true,
-      long long max_memory_bytes = -1  // unlimited
+      int64_t max_memory_bytes = -1  // unlimited
   );
 
   /*! \brief Get the compiled grammar for a JSON schema string. */
@@ -69,7 +77,8 @@ class GrammarCompiler {
       bool any_whitespace = true,
       std::optional<int> indent = std::nullopt,
       std::optional<std::pair<std::string, std::string>> separators = std::nullopt,
-      bool strict_mode = true
+      bool strict_mode = true,
+      std::optional<int> max_whitespace_cnt = std::nullopt
   );
 
   /*! \brief Get the compiled grammar for pure JSON. */
@@ -78,11 +87,14 @@ class GrammarCompiler {
   /*! \brief Get the compiled grammar for a grammar. */
   CompiledGrammar CompileGrammar(const Grammar& grammar);
 
-  /*! \brief Get the compiled grammar for a structural tag. */
-  CompiledGrammar CompileStructuralTag(
-      const std::vector<StructuralTagItem>& tags, const std::vector<std::string>& triggers
+  /*! \brief Get the compiled grammar for a grammar. */
+  CompiledGrammar CompileGrammar(
+      const std::string& ebnf_str, const std::string& root_rule_name = "root"
   );
 
+  /*! \brief Get the compiled grammar for a structural tag. */
+  CompiledGrammar CompileStructuralTag(const std::string& structural_tag_json);
+
   /*! \brief Get the compiled grammar for a regex. */
   CompiledGrammar CompileRegex(const std::string& regex);
 
@@ -90,10 +102,10 @@ class GrammarCompiler {
   void ClearCache();
 
   /*! \brief Return the approximate memory usage of the compiler in bytes. */
-  long long GetCacheSizeBytes() const;
+  int64_t GetCacheSizeBytes() const;
 
   /*! \brief Return the approximate memory usage of the compiler in bytes. -1 means unlimited. */
-  long long CacheLimitBytes() const;
+  int64_t CacheLimitBytes() const;
 
   XGRAMMAR_DEFINE_PIMPL_METHODS(GrammarCompiler);
 };
