@@ -10,9 +10,13 @@
 #include <xgrammar/object.h>
 
 #include <optional>
+#include <stdexcept>
 #include <string>
+#include <variant>
 #include <vector>
 
+#include "xgrammar/exception.h"
+
 namespace xgrammar {
 
 struct StructuralTagItem {
@@ -112,6 +116,7 @@ class Grammar {
       std::optional<int> indent = std::nullopt,
       std::optional<std::pair<std::string, std::string>> separators = std::nullopt,
       bool strict_mode = true,
+      std::optional<int> max_whitespace_cnt = std::nullopt,
       bool print_converted_ebnf = false
   );
 
@@ -124,15 +129,16 @@ class Grammar {
   static Grammar FromRegex(const std::string& regex, bool print_converted_ebnf = false);
 
   /*!
-   * \brief Construct a grammar from a regular expression string.
-   * \param regex The regular expression string.
+   * \brief Construct a grammar from a structural tag string.
+   * \param structural_tag_json The structural tag string.
    */
-  static Grammar FromStructuralTag(
-      const std::vector<StructuralTagItem>& tags, const std::vector<std::string>& triggers
+  static std::variant<Grammar, StructuralTagError> FromStructuralTag(
+      const std::string& structural_tag_json
   );
 
   /*!
    * \brief Get the grammar of standard JSON format. We have built-in support for JSON.
+   * \return The grammar of standard JSON format.
    */
   static Grammar BuiltinJSONGrammar();
 
@@ -152,14 +158,27 @@ class Grammar {
    */
   static Grammar Concat(const std::vector<Grammar>& grammars);
 
-  /*! \brief Print a BNF grammar. */
+  /*!
+   * \brief Print a BNF grammar.
+   * \param os The output stream.
+   * \param grammar The grammar to print.
+   * \return The output stream.
+   */
   friend std::ostream& operator<<(std::ostream& os, const Grammar& grammar);
 
-  /*! \brief Return the serialized JSON string of the grammar. */
+  /*!
+   * \brief Return the serialized JSON string of the grammar.
+   * \return The serialized JSON string.
+   */
   std::string SerializeJSON() const;
 
-  /*! \brief Deserialize a grammar from a JSON string. */
-  static Grammar DeserializeJSON(const std::string& json_string);
+  /*!
+   * \brief Deserialize a grammar from a JSON string.
+   * \param json_string The JSON string to deserialize.
+   * \return If the deserialization is successful, return the grammar. Otherwise, return a runtime
+   * error with the error message.
+   */
+  static std::variant<Grammar, SerializationError> DeserializeJSON(const std::string& json_string);
 
   XGRAMMAR_DEFINE_PIMPL_METHODS(Grammar);
 };
