@@ -8,6 +8,8 @@
 #define XGRAMMAR_XGRAMMAR_H_
 
 #include <xgrammar/compiler.h>
+#include <xgrammar/config.h>
+#include <xgrammar/exception.h>
 #include <xgrammar/grammar.h>
 #include <xgrammar/matcher.h>
 #include <xgrammar/tokenizer_info.h>
