@@ -3,9 +3,16 @@
 from typing import Any, Dict, List, Optional, Tuple, Type, Union, overload
 
 from pydantic import BaseModel
+from typing_extensions import deprecated
 
 from .base import XGRObject, _core
-from .grammar import Grammar, StructuralTagItem, _convert_schema_to_str
+from .grammar import (
+    Grammar,
+    StructuralTagItem,
+    _convert_schema_to_str,
+    _get_structural_tag_str_from_args,
+)
+from .structural_tag import StructuralTag
 from .tokenizer_info import TokenizerInfo
 
 
@@ -37,12 +44,54 @@ class CompiledGrammar(XGRObject):
         return self._handle.memory_size_bytes
 
     def serialize_json(self) -> str:
-        """Serialize the compiled grammar to a JSON string."""
+        """Serialize the compiled grammar to a JSON string. It will serialize the compiled grammar
+        without the tokenizer info, since the tokenizer info is shared by multiple compiled
+        grammars.
+
+        Notes
+        -----
+        The metadata of the tokenizer info is serialized and will be checked when deserializing.
+
+        Returns
+        -------
+        json_string : str
+            The JSON string.
+        """
         return self._handle.serialize_json()
 
     @staticmethod
     def deserialize_json(json_str: str, tokenizer_info: TokenizerInfo) -> "CompiledGrammar":
-        """Deserialize the compiled grammar from a JSON string and tokenizer info."""
+        """Deserialize the compiled grammar from a JSON string and associate it with the specified
+        tokenizer info.
+
+        Notes
+        -----
+        This will check the metadata of the tokenizer info matching the serialized metadata in
+        json_str. If the metadata does not match, a DeserializeFormatError will be raised.
+
+        Parameters
+        ----------
+        json_str : str
+            The JSON string.
+
+        tokenizer_info : TokenizerInfo
+            The tokenizer info.
+
+        Returns
+        -------
+        compiled_grammar : CompiledGrammar
+            The compiled grammar.
+
+        Raises
+        ------
+        InvalidJSONError
+            When the JSON string is invalid.
+        DeserializeFormatError
+            When the JSON string does not follow the serialization format of the grammar, or the
+            tokenizer info metadata does not match.
+        DeserializeVersionError
+            When the __VERSION__ field in the JSON string is not the same as the current version.
+        """
         return CompiledGrammar._create_from_handle(
             _core.CompiledGrammar.deserialize_json(json_str, tokenizer_info._handle)
         )
@@ -100,6 +149,7 @@ class GrammarCompiler(XGRObject):
         indent: Optional[int] = None,
         separators: Optional[Tuple[str, str]] = None,
         strict_mode: bool = True,
+        max_whitespace_cnt: Optional[int] = None,
     ) -> CompiledGrammar:
         """Get CompiledGrammar from the specified JSON schema and format. The indent
         and separators parameters follow the same convention as in json.dumps().
@@ -125,6 +175,12 @@ class GrammarCompiler(XGRObject):
             This helps LLM to generate accurate output in the grammar-guided generation with JSON
             schema.
 
+        max_whitespace_cnt : Optional[int], default: None
+            The maximum number of whitespace characters allowed between elements, such like keys, values, separators and so on.
+            If None, there is no limit on the number of whitespace characters.
+            If specified, it will limit the number of whitespace characters to at most max_whitespace_cnt.
+            It should be a positive integer.
+
         Returns
         -------
         compiled_grammar : CompiledGrammar
@@ -133,7 +189,7 @@ class GrammarCompiler(XGRObject):
         schema_str = _convert_schema_to_str(schema)
         return CompiledGrammar._create_from_handle(
             self._handle.compile_json_schema(
-                schema_str, any_whitespace, indent, separators, strict_mode
+                schema_str, any_whitespace, indent, separators, strict_mode, max_whitespace_cnt
             )
         )
 
@@ -162,28 +218,65 @@ class GrammarCompiler(XGRObject):
         """
         return CompiledGrammar._create_from_handle(self._handle.compile_regex(regex))
 
+    @overload
+    def compile_structural_tag(
+        self, structural_tag: Union[StructuralTag, str, Dict[str, Any]]
+    ) -> CompiledGrammar: ...
+
+    @overload
+    @deprecated(
+        "compile_structural_tag(tags, triggers) is deprecated. Compile structural tag with the "
+        "StructuralTag class instead."
+    )
     def compile_structural_tag(
         self, tags: List[StructuralTagItem], triggers: List[str]
-    ) -> CompiledGrammar:
-        """Compile a grammar from structural tags. See Grammar.from_structural_tag() for more
-        details.
+    ) -> CompiledGrammar: ...
+
+    def compile_structural_tag(self, *args, **kwargs) -> CompiledGrammar:
+        """Compile a grammar from a structural tag. See the Structural Tag Usage in XGrammar
+        documentation for its usage.
+
+        This method supports two calling patterns:
+
+        1. Single structural tag parameter:
+           compile_structural_tag(structural_tag)
+
+        2. Legacy pattern (deprecated):
+           compile_structural_tag(tags, triggers)
 
         Parameters
         ----------
+        structural_tag : Union[StructuralTag, str, Dict[str, Any]]
+            The structural tag either as a StructuralTag object, or a JSON string or a dictionary.
+
         tags : List[StructuralTagItem]
-            The structural tags.
+            (Deprecated) The structural tags. Use StructuralTag class instead.
 
         triggers : List[str]
-            The triggers.
+            (Deprecated) The triggers. Use StructuralTag class instead.
 
         Returns
         -------
         compiled_grammar : CompiledGrammar
-            The compiled grammar.
+            The compiled grammar from the structural tag.
+
+        Raises
+        ------
+        InvalidJSONError
+            When the structural tag is not a valid JSON string.
+        InvalidStructuralTagError
+            When the structural tag is not valid.
+        TypeError
+            When the arguments are invalid.
+
+        Notes
+        -----
+        The legacy pattern compile_structural_tag(tags, triggers) is deprecated. Use the
+        StructuralTag class to construct structural tags instead.
         """
-        tags_tuple = [(tag.begin, _convert_schema_to_str(tag.schema_), tag.end) for tag in tags]
+        structural_tag_str = _get_structural_tag_str_from_args(args, kwargs)
         return CompiledGrammar._create_from_handle(
-            self._handle.compile_structural_tag(tags_tuple, triggers)
+            self._handle.compile_structural_tag(structural_tag_str)
         )
 
     @overload
@@ -223,8 +316,13 @@ class GrammarCompiler(XGRObject):
             The compiled grammar.
         """
         if isinstance(grammar, str):
-            grammar = Grammar.from_ebnf(grammar, root_rule_name=root_rule_name)
-        return CompiledGrammar._create_from_handle(self._handle.compile_grammar(grammar._handle))
+            return CompiledGrammar._create_from_handle(
+                self._handle.compile_grammar(grammar, root_rule_name)
+            )
+        else:
+            return CompiledGrammar._create_from_handle(
+                self._handle.compile_grammar(grammar._handle)
+            )
 
     def clear_cache(self) -> None:
         """Clear all cached compiled grammars."""
