@@ -146,7 +146,7 @@ def test_repetition_range_exact():
     """Test repetition range with exact count {n}."""
     before = """root ::= "a"{3}
 """
-    expected = """root ::= ((("a" "a" "a")))
+    expected = """root ::= (((("a" "a" "a"))))
 """
     grammar = _ebnf_to_grammar_no_normalization(before)
     after = str(grammar)
@@ -157,9 +157,7 @@ def test_repetition_range_min_max():
     """Test repetition range with min and max {n,m}."""
     before = """root ::= "a"{2,4}
 """
-    expected = """root ::= ((("a" "a" root_1)))
-root_1 ::= ("" | ("a" root_2))
-root_2 ::= ("" | "a")
+    expected = """root ::= (((("a" "a") | ("a" "a" "a") | ("a" "a" "a" "a"))))
 """
     grammar = _ebnf_to_grammar_no_normalization(before)
     after = str(grammar)
@@ -170,8 +168,8 @@ def test_repetition_range_min_only():
     """Test repetition range with only min {n,}."""
     before = """root ::= "a"{2,}
 """
-    expected = """root ::= ((("a" "a" root_1)))
-root_1 ::= ("" | ("a" root_1))
+    expected = """root ::= ((("a" "a" root_repeat_inf)))
+root_repeat_inf ::= ("" | ("a" root_repeat_inf))
 """
     grammar = _ebnf_to_grammar_no_normalization(before)
     after = str(grammar)
@@ -222,21 +220,6 @@ def test_unicode_escape():
     assert after == expected
 
 
-def test_tag_dispatch():
-    """Test TagDispatch functionality."""
-    before = """root ::= TagDispatch(("tag1", rule1), ("tag2", rule2))
-rule1 ::= "a"
-rule2 ::= "b"
-"""
-    expected = """root ::= ((TagDispatch(("tag1", rule1), ("tag2", rule2))))
-rule1 ::= (("a"))
-rule2 ::= (("b"))
-"""
-    grammar = _ebnf_to_grammar_no_normalization(before)
-    after = str(grammar)
-    assert after == expected
-
-
 def test_complex_grammar():
     """Test a more complex grammar with multiple features."""
     before = """root ::= expr
@@ -281,11 +264,9 @@ rule1 ::= [a-z]{1,3} (=":")
 rule2 ::= [0-9]+ "." [0-9]*
 """
     expected = """root ::= (("start" root_1 "end"))
-rule1 ::= ((([a-z] rule1_1))) (=((":")))
+rule1 ::= (((([a-z]) | ([a-z] [a-z]) | ([a-z] [a-z] [a-z])))) (=((":")))
 rule2 ::= ((rule2_1 "." [0-9]*))
 root_1 ::= ((((rule1) | (rule2)) root_1) | ((rule1) | (rule2)))
-rule1_1 ::= ("" | ([a-z] rule1_2))
-rule1_2 ::= ("" | [a-z])
 rule2_1 ::= (([0-9] rule2_1) | [0-9])
 """
     grammar = _ebnf_to_grammar_no_normalization(before)
@@ -324,12 +305,11 @@ d ::= ((d_1))
 e ::= (([e]* [f]*) | ([g]*))
 c_1 ::= ("" | ("b" c_1))
 d_1 ::= ("" | (d_1_1 d_1))
-d_1_1 ::= (("bcd") | ("pq"))
+d_1_1 ::= (("b" "c" "d") | ("p" "q"))
 """
 
     grammar = _ebnf_to_grammar_no_normalization(before)
     grammar = GrammarFunctor.structure_normalizer(grammar)
-    grammar = GrammarFunctor.byte_string_fuser(grammar)
     after = str(grammar)
     assert after == expected
 
@@ -342,7 +322,6 @@ rule1 ::= (([abc]* [def]*))
 """
     grammar = _ebnf_to_grammar_no_normalization(before)
     grammar = GrammarFunctor.structure_normalizer(grammar)
-    grammar = GrammarFunctor.byte_string_fuser(grammar)
     after = str(grammar)
     assert after == expected
 
@@ -359,31 +338,33 @@ g ::= "g" {0}
 """
 
     expected = """root ::= ((a b c d e f g))
-a ::= (("a" a_1))
-b ::= ((b_5 b_1))
-c ::= ((c_1))
-d ::= ((d_1))
-e ::= (("ee" e_1))
-f ::= (("fff"))
-g ::= (())
-a_1 ::= ("" | ("a"))
-b_1 ::= ("" | (b_1_1 b_2))
-b_2 ::= ("" | (b_2_1 b_3))
-b_3 ::= ("" | (b_3_1 b_4))
-b_4 ::= ("" | (a) | ("b"))
-c_1 ::= ("" | ("c" c_2))
-c_2 ::= ("" | ("c"))
-d_1 ::= ("" | ("d" d_1))
-e_1 ::= ("" | ("e" e_1))
+a ::= (("a") | ("a" "a"))
+b ::= ((a) | ("b") | (b_1 b_2) | (b_3 b_4 b_5) | (b_6 b_7 b_8 b_9) | (b_10 b_11 b_12 b_13 b_14))
+c ::= ("" | ("c") | ("c" "c"))
+d ::= ((d_repeat_inf))
+e ::= (("e" "e" e_repeat_inf))
+f ::= (("f" "f" "f"))
+g ::= ("")
+d_repeat_inf ::= ("" | ("d" d_repeat_inf))
+e_repeat_inf ::= ("" | ("e" e_repeat_inf))
+b_1 ::= ((a) | ("b"))
+b_2 ::= ((a) | ("b"))
+b_3 ::= ((a) | ("b"))
+b_4 ::= ((a) | ("b"))
 b_5 ::= ((a) | ("b"))
-b_1_1 ::= ((a) | ("b"))
-b_2_1 ::= ((a) | ("b"))
-b_3_1 ::= ((a) | ("b"))
+b_6 ::= ((a) | ("b"))
+b_7 ::= ((a) | ("b"))
+b_8 ::= ((a) | ("b"))
+b_9 ::= ((a) | ("b"))
+b_10 ::= ((a) | ("b"))
+b_11 ::= ((a) | ("b"))
+b_12 ::= ((a) | ("b"))
+b_13 ::= ((a) | ("b"))
+b_14 ::= ((a) | ("b"))
 """
 
     grammar = _ebnf_to_grammar_no_normalization(before)
     grammar = GrammarFunctor.structure_normalizer(grammar)
-    grammar = GrammarFunctor.byte_string_fuser(grammar)
     after = str(grammar)
     assert after == expected
 
@@ -403,7 +384,6 @@ d_choice ::= (("e") | ("d"))
 """
     grammar = _ebnf_to_grammar_no_normalization(before)
     grammar = GrammarFunctor.structure_normalizer(grammar)
-    grammar = GrammarFunctor.byte_string_fuser(grammar)
     after = str(grammar)
     assert after == expected
 
@@ -413,14 +393,13 @@ def test_char():
 rest ::= [a-zA-Z0-9-] [\u0234-\U00000345] [测-试] [\--\]]  rest1
 rest1 ::= "\?\"\'测试あc" "👀" "" [a-a] [b-b]
 """
-    expected = r"""root ::= (([a-z] [A-z] "\u0234\u0345\xff" [\-A-Z] [\-\-] [^a] rest))
+    expected = r"""root ::= (([a-z] [A-z] "\u0234" "\u0345\xff" [\-A-Z] [\-\-] [^a] rest))
 rest ::= (([a-zA-Z0-9\-] [\u0234-\u0345] [\u6d4b-\u8bd5] [\--\]] rest1))
-rest1 ::= (("\?\"\'\u6d4b\u8bd5\u3042c\U0001f440ab"))
+rest1 ::= (("\?\"\'\u6d4b\u8bd5\u3042c" "\U0001f440" "a" "b"))
 """
     # Disable unwrap_nesting_rules to expose the result before unwrapping.
     grammar = _ebnf_to_grammar_no_normalization(before)
     grammar = GrammarFunctor.structure_normalizer(grammar)
-    grammar = GrammarFunctor.byte_string_fuser(grammar)
     after = str(grammar)
     assert after == expected
 
@@ -433,7 +412,7 @@ root::="a"  "b" ("c""d"
 
 "f" | "g"
 """
-    expected = """root ::= (("abcde") | ("f") | ("g"))
+    expected = """root ::= (("a" "b" "c" "d" "e") | ("f") | ("g"))
 """
     grammar = xgr.Grammar.from_ebnf(before)
     after = str(grammar)
@@ -443,8 +422,8 @@ root::="a"  "b" ("c""d"
 def test_nest():
     before = """root::= "a" ("b" | "c" "d") | (("e" "f"))
 """
-    expected = """root ::= (("a" root_1) | ("ef"))
-root_1 ::= (("b") | ("cd"))
+    expected = """root ::= (("a" root_1) | ("e" "f"))
+root_1 ::= (("b") | ("c" "d"))
 """
     grammar = xgr.Grammar.from_ebnf(before)
     after = str(grammar)
@@ -454,7 +433,7 @@ root_1 ::= (("b") | ("cd"))
 def test_empty_parentheses():
     before = """root ::= "a" ( ) "b"
 """
-    expected = """root ::= (("ab"))
+    expected = """root ::= (("a" "b"))
 """
     grammar = xgr.Grammar.from_ebnf(before)
     after = str(grammar)
@@ -492,65 +471,6 @@ rule5 ::= (("") | ("g" rule5 "h"))
     assert after == expected
 
 
-def test_lookahead_assertion_analyzer_tag_dispatch():
-    # tag dispatch disables lookahead assertion detection
-    before = r"""root ::= TagDispatch(("tag1", rule1), ("tag2", rule2), ("tag3", rule3), ("tag4", rule4), ("tag5", rule5))
-rule1 ::= "b"
-rule2 ::= "c"
-rule3 ::= "" | "d" rule3
-rule4 ::= "" | "e" rule4 "f"
-rule5 ::= "" | "g" rule5 "h"
-"""
-    expected = r"""root ::= TagDispatch(("tag1", rule1), ("tag2", rule2), ("tag3", rule3), ("tag4", rule4), ("tag5", rule5))
-rule1 ::= (("b"))
-rule2 ::= (("c"))
-rule3 ::= ("" | ("d" rule3))
-rule4 ::= ("" | ("e" rule4 "f"))
-rule5 ::= ("" | ("g" rule5 "h"))
-"""
-    grammar = _ebnf_to_grammar_no_normalization(before)
-    grammar = GrammarFunctor.structure_normalizer(grammar)
-    grammar = GrammarFunctor.byte_string_fuser(grammar)
-    grammar = GrammarFunctor.lookahead_assertion_analyzer(grammar)
-    after = str(grammar)
-    assert after == expected
-
-
-def test_tag_dispatch_end_to_end():
-    before = """root ::= TagDispatch(("tag1", rule1), ("tag2", rule2), ("tag3", rule3))
-rule1 ::= "a"
-rule2 ::= "b"
-rule3 ::= "c"
-"""
-    expected = """root ::= TagDispatch(("tag1", rule1), ("tag2", rule2), ("tag3", rule3))
-rule1 ::= (("a"))
-rule2 ::= (("b"))
-rule3 ::= (("c"))
-"""
-    grammar = xgr.Grammar.from_ebnf(before)
-    after = str(grammar)
-    assert after == expected
-
-
-def test_tag_dispatch_end_to_end_complex():
-    before = """root ::= TagDispatch(("tag1", rule1), ("tag2", rule2), ("tag3", rule3))
-rule1 ::= ("a" TagDispatch(("tag1", rule2), ("tag2", rule3)) | "zzz")
-rule2 ::= TagDispatch(("tag1", rule2), ("tag2", rule3)) | TagDispatch(("tag3", rule2), ("tag4", rule3))
-rule3 ::= "c"
-"""
-    expected = """root ::= TagDispatch(("tag1", rule1), ("tag2", rule2), ("tag3", rule3))
-rule1 ::= (("a" rule1_1) | ("zzz"))
-rule2 ::= ((rule2_1) | (rule2_2))
-rule3 ::= (("c"))
-rule1_1 ::= TagDispatch(("tag1", rule2), ("tag2", rule3))
-rule2_1 ::= TagDispatch(("tag1", rule2), ("tag2", rule3))
-rule2_2 ::= TagDispatch(("tag3", rule2), ("tag4", rule3))
-"""
-    grammar = xgr.Grammar.from_ebnf(before)
-    after = str(grammar)
-    assert after == expected
-
-
 def test_flatten():
     before = """root ::= or_test sequence_test nested_test empty_test
 or_test ::= ([a] | "b") | "de" | "" | or_test | [^a-z]
@@ -561,15 +481,14 @@ empty_test ::= "d" | (("" | "" "") "" | "a" "") | ("" ("" | "")) "" ""
 """
     expected = """root ::= ((or_test sequence_test nested_test empty_test))
 or_test ::= ("" | ("a") | ("b") | ("de") | (or_test) | ([^a-z]))
-sequence_test ::= (("aab" sequence_test_1 "de" sequence_test))
-nested_test ::= (("abcd") | ("a") | ("b") | ("c") | (nested_rest))
-nested_rest ::= (("a") | ("bc") | ("d") | ("ef") | ("g"))
+sequence_test ::= (("a" "a" "b" sequence_test_1 "d" "e" sequence_test))
+nested_test ::= (("a" "b" "c" "d") | ("a") | ("b") | ("c") | (nested_rest))
+nested_rest ::= (("a") | ("b" "c") | ("d") | ("e" "f") | ("g"))
 empty_test ::= ("" | ("d") | ("a"))
 sequence_test_1 ::= (("c") | ("d"))
 """
     grammar = _ebnf_to_grammar_no_normalization(before)
     grammar = GrammarFunctor.structure_normalizer(grammar)
-    grammar = GrammarFunctor.byte_string_fuser(grammar)
     after = str(grammar)
     assert after == expected
 
@@ -760,6 +679,7 @@ sign ::= ("" | ("+") | ("-"))
 """
 
     grammar = xgr.Grammar.from_ebnf(before)
+    grammar = GrammarFunctor.grammar_optimizer(grammar)
     after = str(grammar)
     assert after == expected
 
@@ -782,21 +702,6 @@ d_1 ::= ("" | ("d"))
     assert output_string_1 == output_string_2
 
 
-def test_e2e_tag_dispatch_roundtrip():
-    """Checks the printed result can be parsed, and the parsing-printing process is idempotent."""
-    before = r"""root ::= TagDispatch(("tag1", rule1), ("tag2", rule2), ("tag3", rule3))
-rule1 ::= (("a"))
-rule2 ::= (("b"))
-rule3 ::= (("c"))
-"""
-    grammar_1 = xgr.Grammar.from_ebnf(before)
-    output_string_1 = str(grammar_1)
-    grammar_2 = xgr.Grammar.from_ebnf(output_string_1)
-    output_string_2 = str(grammar_2)
-    assert before == output_string_1
-    assert output_string_1 == output_string_2
-
-
 ebnf_str__expected_error_regex__test_lexer_parser_errors = [
     (r'root ::= "a" "', 'EBNF lexer error at line 1, column 15: Expect " in string literal'),
     (
@@ -846,35 +751,6 @@ def test_end_to_end_errors(ebnf_str: str, expected_error_regex: Optional[str]):
         xgr.Grammar.from_ebnf(ebnf_str)
 
 
-ebnf_str__expected_error_regex__test_tag_dispatch_parser_errors = [
-    (
-        'root ::= TagDispatch(("", rule1))\nrule1 ::= "a"',
-        "EBNF parser error at line 1, column 21: Tag must be a non-empty string literal",
-    ),
-    (
-        'root ::= TagDispatch(("tag1", undefined_rule))',
-        'EBNF parser error at line 1, column 21: Rule "undefined_rule" is not defined',
-    ),
-    (
-        'root ::= TagDispatch("tag1", rule1)',
-        "EBNF parser error at line 1, column 21: Each tag dispatch element must be a tuple",
-    ),
-    (
-        'root ::= TagDispatch(("tag1" rule1))',
-        "EBNF parser error at line 1, column 30: Expect , or \\) in tuple",
-    ),
-]
-
-
-@pytest.mark.parametrize(
-    "ebnf_str, expected_error_regex",
-    ebnf_str__expected_error_regex__test_tag_dispatch_parser_errors,
-)
-def test_tag_dispatch_parser_errors(ebnf_str: str, expected_error_regex: Optional[str]):
-    with pytest.raises(RuntimeError, match=expected_error_regex):
-        _ebnf_to_grammar_no_normalization(ebnf_str)
-
-
 def test_error_consecutive_quantifiers():
     grammar_str = """root ::= "a"{1,3}{1,3}
 """
@@ -898,5 +774,24 @@ def test_error_consecutive_quantifiers():
         xgr.Grammar.from_ebnf(grammar_str)
 
 
+def test_repetition_normalizer():
+    """Test the repetition normalizer. If the context is nullable, then the min repetition time will be reduced to 0."""
+    before = "root ::= ([0-9]*){100, 1000}"
+    expected_grammar = r"""root ::= ((root_repeat_1{0, 985} [0-9]* [0-9]* [0-9]* [0-9]* [0-9]* [0-9]* [0-9]* [0-9]* [0-9]* [0-9]* [0-9]* [0-9]* [0-9]* [0-9]* [0-9]*))
+root_repeat_1 ::= (([0-9]*)) (=([0-9]* [0-9]* [0-9]* [0-9]* [0-9]* [0-9]* [0-9]* [0-9]* [0-9]* [0-9]* [0-9]* [0-9]* [0-9]* [0-9]* [0-9]*))
+"""
+    grammar = xgr.Grammar.from_ebnf(before)
+    grammar = GrammarFunctor.grammar_optimizer(grammar)
+    assert expected_grammar == str(grammar)
+
+    before = "root ::= ([0-9]){100, 1000}"
+    expected_grammar = r"""root ::= ((root_repeat_1{85, 985} [0-9] [0-9] [0-9] [0-9] [0-9] [0-9] [0-9] [0-9] [0-9] [0-9] [0-9] [0-9] [0-9] [0-9] [0-9]))
+root_repeat_1 ::= (([0-9])) (=([0-9] [0-9] [0-9] [0-9] [0-9] [0-9] [0-9] [0-9] [0-9] [0-9] [0-9] [0-9] [0-9] [0-9] [0-9]))
+"""
+    grammar = xgr.Grammar.from_ebnf(before)
+    grammar = GrammarFunctor.grammar_optimizer(grammar)
+    assert expected_grammar == str(grammar)
+
+
 if __name__ == "__main__":
     pytest.main(sys.argv)
